from desssign.loads.wind.enums import TerrainCategory
from desssign.loads.wind.enums import WindZone

# Fundamental value of basic wind velocity (v_b0) in m/s.
FUNDAMENTAL_VALUE_OF_BASIC_WIND_VELOCITY: dict[str, float] = {
    WindZone.I: 22.5,
    WindZone.II: 25.0,
    WindZone.III: 27.5,
    WindZone.IV: 30.0,
    WindZone.V: 36.0,
}

# Directional factor (c_dir)
DIRECTIONAL_FACTOR: float = 1.0

# Seasonal factor (c_season)
SEASONAL_FACTOR: float = 1.0

# Density of air (rho_air) in kg/m³.
AIR_DENSITY: float = 1.25

# Roughness length (z_0) in m, according to EN 1991-1-4, table 4.1.
ROUGHNESS_LENGTH: dict[str, float] = {
    TerrainCategory.O: 0.003,
    TerrainCategory.I: 0.01,
    TerrainCategory.II: 0.05,
    TerrainCategory.III: 0.3,
    TerrainCategory.IV: 1.0,
}

# Minimum height (z_min) in m, according to EN 1991-1-4, table 4.1.
MINIMUM_HEIGHT: dict[str, float] = {
    TerrainCategory.O: 1.0,
    TerrainCategory.I: 1.0,
    TerrainCategory.II: 2.0,
    TerrainCategory.III: 5.0,
    TerrainCategory.IV: 10.0,
}



from __future__ import annotations

from math import log

from desssign.loads.wind.constants import AIR_DENSITY
from desssign.loads.wind.constants import DIRECTIONAL_FACTOR
from desssign.loads.wind.constants import FUNDAMENTAL_VALUE_OF_BASIC_WIND_VELOCITY
from desssign.loads.wind.constants import MINIMUM_HEIGHT
from desssign.loads.wind.constants import ROUGHNESS_LENGTH
from desssign.loads.wind.constants import SEASONAL_FACTOR
from desssign.loads.wind.enums import TerrainCategory
from desssign.loads.wind.enums import WindZone


class WindLoad:
    """
    Class representing wind load according to EN 1991-1-4.

    :ivar zone: Wind zone.
    :ivar c_dir: Directional factor.
    :ivar c_season: Seasonal factor.
    :ivar rho_air: Density of air in kg/m³.
    """

    def __init__(
        self,
        zone: str | WindZone,
        terrain_category: str | TerrainCategory,
        z_e: float,
        area: float = 10.0,
        c_dir: float = DIRECTIONAL_FACTOR,
        c_season: float = SEASONAL_FACTOR,
        rho_air: float = AIR_DENSITY,
    ) -> None:
        self.zone = zone
        self.terrain_category = terrain_category
        self.z_e = z_e
        self.area = area

        self.c_dir = c_dir
        self.c_season = c_season
        self.rho_air = rho_air

    @property
    def v_b0(self) -> float:
        """
        Fundamental value of basic wind velocity in m/s according to EN 1991-1-4.

        :return: Fundamental value of basic wind velocity in m/s.
        """
        return FUNDAMENTAL_VALUE_OF_BASIC_WIND_VELOCITY[self.zone]

    @property
    def v_b(self) -> float:
        """
        Basic wind velocity in m/s according to EN 1991-1-4, eq. (4.1).

        The basic wind velocity is defined as function of wind direction and time of year
        at 10 m above the ground of terrain category II.

        :return: Basic wind velocity in m/s.
        """
        return self.c_dir * self.c_season * self.v_b0

    @property
    def q_b(self) -> float:
        """
        Basic velocity pressure in N/m² according to EN 1991-1-4, eq. (4.10).

        :return: Basic velocity pressure in N/m².
        """
        return 0.5 * self.rho_air * self.v_b**2

    @property
    def z_0(self) -> float:
        """Roughness length in m according to EN 1991-1-4, table 4.1."""
        return ROUGHNESS_LENGTH[self.terrain_category]

    @property
    def z_min(self) -> float:
        """Minimum height in m according to EN 1991-1-4, table 4.1."""
        return MINIMUM_HEIGHT[self.terrain_category]

    @property
    def z_max(self) -> float:
        """Maximum height in m according to EN 1991-1-4."""
        return 200.0

    @property
    def k_r(self) -> float:
        """
        Terrain factor.

        According to EN 1991-1-4, 4.3.2, eq. (4.5).
        """
        z_0II: float = ROUGHNESS_LENGTH[TerrainCategory.II]

        return 0.19 * (self.z_0 / z_0II) ** 0.07  # type: ignore[no-any-return]

    @property
    def k_l(self) -> float:
        """
        Turbulence factor.

        According to EN 1991-1-4, 4.4.1(1), note 2.
        """
        return 1.0

    @property
    def sigma_v(self) -> float:
        """Standard deviation of the turbulence in m/s according to EN 1991-1-4, 4.4(1), eq. (4.6)."""
        return self.k_r * self.v_b * self.k_l

    @property
    def c_o(self) -> float:
        """Orography factor."""
        return 1.0

    @property
    def c_r(self) -> float:
        """Roughness factor according to EN 1991-1-4, 4.3.2(1), eq. (4.4)."""
        if self.z_min <= self.z_e <= self.z_max:
            return self.k_r * log(self.z_e / self.z_0)
        elif self.z_e < self.z_min:
            return self.k_r * log(self.z_min / self.z_0)
        else:
            raise ValueError("Height above the maximum height is not supported.")

    @property
    def v_m(self) -> float:
        """Mean wind velocity in m/s according to EN 1991-1-4, 4.3.1 (1), eq. (4.3)."""
        return self.c_r * self.c_o * self.v_b

    @property
    def I_v(self) -> float:
        """Turbulence intensity according to EN 1991-1-4, 4.4.1(1), eq. (4.7)."""
        if self.z_min <= self.z_e <= self.z_max:
            return self.sigma_v / self.v_m
        elif self.z_e < self.z_min:
            return self.sigma_v / self.v_m
        else:
            raise ValueError("Height above the maximum height is not supported.")

    @property
    def q_p(self) -> float:
        """Peak velocity pressure in N/m² according to EN 1991-1-4, 4.5.1(1), eq. (4.8)."""
        return (1 + 7 * self.I_v) * 1 / 2 * self.rho_air * self.v_m**2

    @property
    def c_e(self) -> float:
        """Exposure factor according to EN 1991-1-4, 4.5.1(1), eq. (4.9)."""
        return self.q_p / self.q_b



from __future__ import annotations

import math
from typing import TYPE_CHECKING

from framesss.pre.member_1d import Member1D

from desssign.wood.design_checks.member_1d_checks import WoodMember1DChecks

if TYPE_CHECKING:
    import numpy as np
    import numpy.typing as npt
    from framesss.enums import BeamConnection
    from framesss.enums import Element1DType
    from framesss.fea.analysis.analysis import Analysis
    from framesss.fea.node import Node

    from desssign.loads.load_case_combination import DesignLoadCaseCombination
    from desssign.wood.wood_section import WoodRectangularSection


class WoodMember1D(Member1D):
    """
    Class for 1D wood member.

    Inherits from `Member1D` class and adds specific properties of wood.

    param label: A user-defined identifier for the member.
    :param element_type: Specifies the type of the element ('navier', 'timoshenko').
    :param nodes: The nodes at the ends of the member.
    :param section: The cross-section of the member.
    :param hinges: Defines the type of connections at the start and the end of the  member
                   (e.g., fixed, hinged, or semirigid) to model the rotational stiffness accurately.
    :param auxiliary_vector_xy_plane: An auxiliary vector in local xy-plane that defines the local
                                      coordinate system of the member.
    :param analysis: The :class:`Analysis` object.
    """

    section: WoodRectangularSection  # Explicit type annotation, so that mypy can check the type

    def __init__(
        self,
        label: str,
        element_type: str | Element1DType,
        nodes: list[Node],
        section: WoodRectangularSection,
        hinges: list[str] | list[BeamConnection],
        auxiliary_vector_xy_plane: npt.NDArray[np.float64],
        analysis: Analysis,
    ) -> None:
        """Init the WoodMember1D object."""
        super().__init__(
            label=label,
            element_type=element_type,
            nodes=nodes,
            section=section,
            hinges=hinges,
            auxiliary_vector_xy_plane=auxiliary_vector_xy_plane,
            analysis=analysis,
        )

        self.design_checks = WoodMember1DChecks(self)

        self.is_prevented_lateral_displacement_at_compressive_edge = False
        self.is_prevented_torsional_rotation_at_supports = False

    @property
    def L_cr_y(self) -> float:
        """Critical length for buckling in y-axis."""
        # TODO: Implement the critical length for buckling in y-axis
        return float(self.length)

    @property
    def L_cr_z(self) -> float:
        """Critical length for buckling in z-axis."""
        # TODO: Implement the critical length for buckling in z-axis
        return float(self.length)

    @property
    def lambda_y(self) -> float:
        """Slenderness ratio in y-axis."""
        return self.L_cr_y / self.section.i_y

    @property
    def lambda_z(self) -> float:
        """Slenderness ratio in z-axis."""
        return self.L_cr_z / self.section.i_z

    @property
    def lambda_rel_y(self) -> float:
        """
        Relative slenderness ratio in y-axis.

        EN 1995-1-1, 6.3.2(1), eq. (6.21)
        """
        return (
            self.lambda_y
            / math.pi
            * math.sqrt(self.section.material.f_c0k / self.section.material.E_m0k)
        )

    @property
    def lambda_rel_z(self) -> float:
        """
        Relative slenderness ratio in z-axis.

        EN 1995-1-1, 6.3.2(1), eq. (6.22)
        """
        return (
            self.lambda_z
            / math.pi
            * math.sqrt(self.section.material.f_c0k / self.section.material.E_m0k)
        )

    @property
    def k_y(self) -> float:
        """
        Instability factor.

        EN 1995-1-1, 6.3.2(3), eq. (6.27)
        """
        beta_c = self.section.material.beta_c
        lambda_rel_y = self.lambda_rel_y
        return 0.5 * (1 + beta_c * (lambda_rel_y - 0.3) + lambda_rel_y**2)

    @property
    def k_z(self) -> float:
        """
        Instability factor.

        EN 1995-1-1, 6.3.2(3), eq. (6.28)
        """
        beta_c = self.section.material.beta_c
        lambda_rel_z = self.lambda_rel_z
        return 0.5 * (1 + beta_c * (lambda_rel_z - 0.3) + lambda_rel_z**2)

    @property
    def k_cy(self) -> float:
        """
        Instability factor.

        EN 1995-1-1, 6.3.2(3), eq. (6.25)
        """
        return 1.0 / (self.k_y + math.sqrt(self.k_y**2 - self.lambda_rel_y**2))

    @property
    def k_cz(self) -> float:
        """
        Instability factor.

        EN 1995-1-1, 6.3.2(3), eq. (6.26)
        """
        return 1.0 / (self.k_z + math.sqrt(self.k_z**2 - self.lambda_rel_z**2))

    @property
    def l_ef(self) -> float:
        """
        Effective length of the member subjected to bending.

        EN 1995-1-1, 6.3.3(3), Table 6.1
        """
        # TODO: Implement the effective length of the member subjected to bending
        return float(0.9 * self.length + 2 * self.section.height_z)

    @property
    def sigma_m_crit(self) -> float:
        """
        Critical bending stress.

        EN 1995-1-1, 6.3.3(3), eq. (6.32)
        """
        b = self.section.height_y
        E = self.section.material.E_m0k
        h = self.section.height_z
        l_ef = self.l_ef
        return 0.78 * b**2 * E / (h * l_ef)

    @property
    def lambda_rel_m(self) -> float:
        """
        Relative slenderness for bending.

        EN 1995-1-1, 6.3.3(2), eq. (6.30)
        """
        return math.sqrt(self.section.material.f_mk / self.sigma_m_crit)

    @property
    def k_crit(self) -> float:
        """
        Factor used for lateral buckling.

        EN 1995-1-1, 6.3.3(4), eq. (6.34)
        EN 1995-1-1, 6.3.3(5)
        """
        if (
            self.is_prevented_lateral_displacement_at_compressive_edge
            and self.is_prevented_torsional_rotation_at_supports
        ):
            return 1.0

        if self.lambda_rel_m <= 0.75:
            return 1.0
        elif 0.75 < self.lambda_rel_m <= 1.4:
            return 1.56 - 0.75 * self.lambda_rel_m
        else:
            return 1 / self.lambda_rel_m**2

    def perform_uls_checks(
        self, load_combinations: list[DesignLoadCaseCombination]
    ) -> None:
        """Perform the design checks."""
        self.design_checks.perform_uls_checks(load_combinations)



from desssign.loads.snow.enums import SnowZone
from desssign.loads.snow.enums import Topography

EXPOSURE_COEFFICIENTS = {
    Topography.WINDSWEPT: 0.8,
    Topography.NORMAL: 1.0,
    Topography.SHELTERED: 1.2,
}

SNOW_LOAD_ON_THE_GROUND = {
    SnowZone.I: 0.7,
    SnowZone.II: 1.0,
    SnowZone.III: 1.5,
    SnowZone.IV: 2.0,
    SnowZone.V: 2.5,
    SnowZone.VI: 3.0,
    SnowZone.VII: 4.0,
    SnowZone.VIII: 0.0,
}



from __future__ import annotations

from typing import cast

from framesss.pre.cases import LoadCase
from framesss.pre.cases import LoadCaseCombination
from framesss.pre.cases import NonlinearLoadCaseCombination

from desssign.loads.enums import LOAD_DURATION_MAPPING
from desssign.loads.enums import LimitState
from desssign.loads.enums import LoadDurationClass
from desssign.loads.enums import SLSCombination
from desssign.loads.enums import ULSAlternativeCombination
from desssign.loads.enums import ULSCombination
from desssign.loads.load_case import DesignLoadCase
from desssign.loads.load_combination_generator.generate_combinations import (
    generate_combination,
)


class DesignLoadCaseCombination(LoadCaseCombination):
    """
    Represent a combination of load cases in the design process.

    :param label: The label of the load case combination.
    :param limit_state: The limit state of the combination group. Either 'ULS' or 'SLS'.
    :param combination_type: The type of the combination group. For ULS: basic, alternative or accidental,
                             for SLS: characteristic, frequent or quasi-permanent.
    :param permanent_cases: A list of permanent load cases.
    :param leading_variable_case: The leading variable load case.
    :param other_variable_cases: A list of other variable load cases.
    :param alternative_combination: Specifier for the used equation, only required for ULS alternative combinations.
                                    Either '6.10a' or '6.10b'.
    :param description: A description of the load case combination.
    :ivar combination_key: The key of the combination.
    """

    def __init__(
        self,
        label: str,
        limit_state: str | LimitState,
        combination_type: str | SLSCombination | ULSCombination,
        permanent_cases: list[DesignLoadCase],
        leading_variable_case: DesignLoadCase | None,
        other_variable_cases: list[DesignLoadCase],
        description: str = "",
        alternative_combination: str | ULSAlternativeCombination | None = None,
    ) -> None:
        """Initialize the DesignLoadCaseCombination class."""
        self.limit_state = LimitState(limit_state)

        if self.limit_state == LimitState.ULS:
            self.combination_type = ULSCombination(combination_type)
        elif self.limit_state == LimitState.SLS:
            self.combination_type = SLSCombination(combination_type)
        else:
            raise ValueError(
                f"Can't set combination type: '{combination_type}' to limit state: '{limit_state}'."
            )

        if combination_type == ULSCombination.ALTERNATIVE:
            if alternative_combination is None:
                raise ValueError(
                    "Alternative combination requires an 'alternative_combination'."
                )

        self.alternative_combination = (
            ULSAlternativeCombination(alternative_combination)
            if alternative_combination
            else None
        )

        self.permanent_cases = permanent_cases
        self.leading_variable_case = leading_variable_case
        self.other_variable_cases = other_variable_cases

        load_cases, self.combination_key = generate_combination(
            permanent_cases=self.permanent_cases,
            leading_variable_case=self.leading_variable_case,
            other_variable_cases=self.other_variable_cases,
            combination=self.combination_type,
            alternative_combination=self.alternative_combination,
        )

        self.description = description

        super().__init__(label, cast(dict[LoadCase, float], load_cases))

    def _get_combination(self) -> tuple[dict[DesignLoadCase, float], str]:
        """Return the load cases combination and the combination key."""
        return generate_combination(
            permanent_cases=self.permanent_cases,
            leading_variable_case=self.leading_variable_case,
            other_variable_cases=self.other_variable_cases,
            combination=self.combination_type,
            alternative_combination=self.alternative_combination,
        )

    @property
    def load_duration_class(self) -> LoadDurationClass:
        """
        Return the load duration class of the combination.

        EN 1995-1-1, 3.1.3(2):
            If a load combination consists of actions belonging to different load-duration classes a value
            of `k_mod` should be chosen which corresponds to the action with the shortest duration, e.g. for a
            combination of dead load and a short-term load, a value of k_mod corresponding to the short-term
            load should be used.
        """
        # Get the minimum value of all duration classes
        min_duration_value = min(
            [
                LOAD_DURATION_MAPPING[cast(DesignLoadCase, case).load_duration_class]
                for case in self.load_cases
            ]
        )

        # Find the LoadDurationClass corresponding to the minimum value
        for duration_class, value in LOAD_DURATION_MAPPING.items():
            if value == min_duration_value:
                return LoadDurationClass(duration_class)

        # If no matching LoadDurationClass is found, raise an exception
        raise ValueError(
            f"Can't find the load duration class with value: '̈́{min_duration_value}'."
        )


class DesignNonlinearLoadCaseCombination(NonlinearLoadCaseCombination):
    """
    Represent a combination of load cases in the design process.

    :param label: The label of the load case combination.
    :param limit_state: The limit state of the combination group. Either 'ULS' or 'SLS'.
    :param combination_type: The type of the combination group. For ULS: basic, alternative or accidental,
                             for SLS: characteristic, frequent or quasi-permanent.
    :param permanent_cases: A list of permanent load cases.
    :param leading_variable_case: The leading variable load case.
    :param other_variable_cases: A list of other variable load cases.
    :param alternative_combination: Specifier for the used equation, only required for ULS alternative combinations.
                                    Either '6.10a' or '6.10b'.
    :param description: A description of the load case combination.
    :ivar combination_key: The key of the combination.
    """

    def __init__(
        self,
        label: str,
        limit_state: str | LimitState,
        combination_type: str | SLSCombination | ULSCombination,
        permanent_cases: list[DesignLoadCase],
        leading_variable_case: DesignLoadCase | None,
        other_variable_cases: list[DesignLoadCase],
        description: str = "",
        alternative_combination: str | ULSAlternativeCombination | None = None,
    ) -> None:
        """Initialize the NonlinearDesignLoadCaseCombination class."""
        self.limit_state = LimitState(limit_state)

        if self.limit_state == LimitState.ULS:
            self.combination_type = ULSCombination(combination_type)
        elif self.limit_state == LimitState.SLS:
            self.combination_type = SLSCombination(combination_type)
        else:
            raise ValueError(
                f"Can't set combination type: '{combination_type}' to limit state: '{limit_state}'."
            )

        if combination_type == ULSCombination.ALTERNATIVE:
            if alternative_combination is None:
                raise ValueError(
                    "Alternative combination requires an 'alternative_combination'."
                )

        self.alternative_combination = (
            ULSAlternativeCombination(alternative_combination)
            if alternative_combination
            else None
        )

        self.permanent_cases = permanent_cases
        self.leading_variable_case = leading_variable_case
        self.other_variable_cases = other_variable_cases

        load_cases, self.combination_key = generate_combination(
            permanent_cases=self.permanent_cases,
            leading_variable_case=self.leading_variable_case,
            other_variable_cases=self.other_variable_cases,
            combination=self.combination_type,
            alternative_combination=self.alternative_combination,
        )

        self.description = description

        super().__init__(label, cast(dict[LoadCase, float], load_cases))

    def _get_combination(self) -> tuple[dict[DesignLoadCase, float], str]:
        """Return the load cases combination and the combination key."""
        return generate_combination(
            permanent_cases=self.permanent_cases,
            leading_variable_case=self.leading_variable_case,
            other_variable_cases=self.other_variable_cases,
            combination=self.combination_type,
            alternative_combination=self.alternative_combination,
        )

    @property
    def load_duration_class(self) -> LoadDurationClass:
        """
        Return the load duration class of the combination.

        EN 1995-1-1, 3.1.3(2):
            If a load combination consists of actions belonging to different load-duration classes a value
            of `k_mod` should be chosen which corresponds to the action with the shortest duration, e.g. for a
            combination of dead load and a short-term load, a value of k_mod corresponding to the short-term
            load should be used.
        """
        # Get the minimum value of all duration classes
        min_duration_value = min(
            [
                LOAD_DURATION_MAPPING[cast(DesignLoadCase, case).load_duration_class]
                for case in self.load_cases
            ]
        )

        # Find the LoadDurationClass corresponding to the minimum value
        for duration_class, value in LOAD_DURATION_MAPPING.items():
            if value == min_duration_value:
                return LoadDurationClass(duration_class)

        # If no matching LoadDurationClass is found, raise an exception
        raise ValueError(
            f"Can't find the load duration class with value: '̈́{min_duration_value}'."
        )



from framesss.enums import CaseInsensitiveStrEnum


class CheckResult(CaseInsensitiveStrEnum):
    """
    Enum for design results.

    :cvar PASS: Design check passed.
    :cvar FAIL: Design check failed.
    """

    PASS = "pass"
    FAIL = "fail"



from __future__ import annotations

import math
from typing import TYPE_CHECKING

from framesss.pre.section import RectangularSection

from desssign.wood.enums import WoodType

if TYPE_CHECKING:
    from desssign.wood.wood_material import WoodMaterial


class WoodRectangularSection(RectangularSection):
    """
    Class for wood rectangular section.

    Inherits from `RectangularSection` class and adds specific properties of wood.

    :param label: Label of the section.
    :param b: Width of the section.
    :param h: Height of the section.
    :param material: Wood material of the section.
    """

    material: WoodMaterial  # Explicit type annotation, so that mypy can check the type

    def __init__(
        self,
        label: str,
        b: float,
        h: float,
        material: WoodMaterial,
    ) -> None:
        super().__init__(label, b, h, material)

    @property
    def i_y(self) -> float:
        """Radius of gyration to the y-axis."""
        return math.sqrt(self.inertia_y / self.area_y)

    @property
    def i_z(self) -> float:
        """Radius of gyration to the z-axis."""
        return math.sqrt(self.inertia_z / self.area_z)

    @property
    def W_y(self) -> float:
        """Elastic section modulus to the y-axis."""
        return self.inertia_y / (0.5 * self.height_z)

    @property
    def W_z(self) -> float:
        """Elastic section modulus to the z-axis."""
        return self.inertia_z / (0.5 * self.height_y)

    @property
    def k_m(self) -> float:
        """
        Factor considering re-distribution of bending stresses in a cross-section.

        EN 1995-1-1, 6.1.6(2)
        """
        if (
            self.material.wood_type == WoodType.SOLID_TIMBER
            or self.material.wood_type == WoodType.GLUED_LAMINATED_TIMBER
            or self.material.wood_type == WoodType.LVL
        ):
            return 0.7
        else:
            return 1.0

    @property
    def k_cr(self) -> float:
        """
        Crack factor for shear resistance.

        EN 1995-1-1, 6.1.7 (2), eq. (6.13a)
        """
        if self.material.wood_type == WoodType.SOLID_TIMBER:
            return 0.67
        elif self.material.wood_type == WoodType.GLUED_LAMINATED_TIMBER:
            return 0.67
        else:
            return 1.0

    def get_k_h(self, type_of_stress: str) -> float:
        """
        Depth factor.

        EN 1995-1-1, 3.2(3):
            For rectangular solid timber with a characteristic timber density `rho_k` <= 700 kg/m3,
            the reference depth in bending or width (maximum cross-sectional dimension) in tension
            is 150 mm. For depths in bending or widths in tension of solid timber less than 150 mm
            the characteristic values for `f_mk` and `f_t0k` may be increased by the factor `k_h`.

        :param type_of_stress: Type of stress, either 'bending' or 'tension'.
        """
        if self.material.rho_k > 700:
            return 1.0

        if type_of_stress.lower() == "bending":
            h = self.height_z
        elif type_of_stress.lower() == "tension":
            h = max(self.height_z, self.height_y)
        else:
            raise AttributeError(
                f"Invalid type of stress: '{type_of_stress}'. Use either 'bending' or 'tension'."
            )

        if h <= 0:
            raise ValueError(
                "Height or width of the section must be greater than zero."
            )

        return float(min((150 / h) ** 0.2, 1.3))



from framesss.enums import CaseInsensitiveStrEnum


class TerrainCategory(CaseInsensitiveStrEnum):
    """Enumeration of terrain categories according to EN 1991-1-4."""

    O = "O"
    I = "I"
    II = "II"
    III = "III"
    IV = "IV"


class WindZone(CaseInsensitiveStrEnum):
    """Enumeration of wind zones according to ČSN EN 1991-1-4."""

    I = "I"
    II = "II"
    III = "III"
    IV = "IV"
    V = "V"


class RoofProfile(CaseInsensitiveStrEnum):
    """Enumeration of roof profiles according to EN 1991-1-4."""

    FLAT = "flat"
    MONOPITCHED = "monopitched"
    DUOPITCHED = "duopitched"


class FlatRoofType(CaseInsensitiveStrEnum):
    """Enumeration of flat roof types according to EN 1991-1-4."""

    SHARP_EAVES = "sharp eaves"
    WITH_PARAPETS = "with parapets"
    # CURVED_EAVES = "curved eaves"
    # MANSARD_EAVES = "mansard eaves"



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from desssign.common.design_check import Member1DChecks
from desssign.concrete.design_checks.design_check import BendingCheck

if TYPE_CHECKING:
    import numpy.typing as npt

    from desssign.loads.load_case_combination import DesignLoadCaseCombination
    from desssign.concrete.concrete_member import ConcreteMember1D


class ConcreteMember1DChecks(Member1DChecks):
    """
    Class for performing design checks on 1D members.

    :param member: The 1D concrete member.
    """

    member: (
        ConcreteMember1D  # Explicit type annotation, so that mypy can check the type
    )

    def __init__(self, member: ConcreteMember1D):
        super().__init__(member=member)

        self.bending_check: dict[DesignLoadCaseCombination, BendingCheck] = {}

    @property
    def max_usage(self) -> float:
        """Maximum usage of the material."""
        max_usages = [check.max_usage for check in (*self.bending_check.values(),)]
        return max(max_usages)

    def perform_uls_checks(
        self,
        load_case_combinations: list[DesignLoadCaseCombination],
    ) -> None:
        self.perform_bending_checks(load_case_combinations)

    def perform_bending_checks(
        self,
        load_case_combinations: list[DesignLoadCaseCombination],
    ) -> None:
        for combination in load_case_combinations:
            axial, _, _, _, bending_y, _ = self.get_internal_forces(combination)

            self.bending_check[combination] = BendingCheck(
                m_ed=bending_y,
                m_rd_positive=self.member.section.m_rd_positive,
                m_rd_negative=self.member.section.m_rd_negative,
            )



from __future__ import annotations

from desssign.loads.enums import LoadDurationClass
from desssign.wood.enums import JointType
from desssign.wood.enums import ServiceClass
from desssign.wood.enums import WoodType

PARTIAL_FACTORS = {
    WoodType.SOLID_TIMBER: 1.3,
    WoodType.GLUED_LAMINATED_TIMBER: 1.25,
    WoodType.LVL: 1.2,
    WoodType.PLYWOOD: 1.2,
    WoodType.OSB: 1.2,
    WoodType.PARTICLEBOARD: 1.3,
    WoodType.HARD_FIBREBOARD: 1.3,
    WoodType.MEDIUM_FIBREBOARD: 1.3,
    WoodType.MDF_FIBREBOARD: 1.3,
    WoodType.SOFT_FIBREBOARD: 1.3,
    JointType.CONNECTION: 1.3,
    JointType.PUNCHED_METAL_PLATE_FASTENERS: 1.25,
}


def get_partial_factor(material: str | WoodType | JointType) -> float:
    """
    Partial factor according to EN 1995-1-1, table 2.5.

    :param material: Material or joint type.
    :return: Partial factor.
    """
    return PARTIAL_FACTORS[material]


MODIFICATION_FACTORS = {
    WoodType.SOLID_TIMBER: {
        ServiceClass.SC1: {
            LoadDurationClass.PERMANENT: 0.6,
            LoadDurationClass.LONG_TERM: 0.7,
            LoadDurationClass.MEDIUM_TERM: 0.8,
            LoadDurationClass.SHORT_TERM: 0.9,
            LoadDurationClass.INSTANTANEOUS: 1.1,
        },
        ServiceClass.SC2: {
            LoadDurationClass.PERMANENT: 0.6,
            LoadDurationClass.LONG_TERM: 0.7,
            LoadDurationClass.MEDIUM_TERM: 0.8,
            LoadDurationClass.SHORT_TERM: 0.9,
            LoadDurationClass.INSTANTANEOUS: 1.1,
        },
        ServiceClass.SC3: {
            LoadDurationClass.PERMANENT: 0.5,
            LoadDurationClass.LONG_TERM: 0.55,
            LoadDurationClass.MEDIUM_TERM: 0.65,
            LoadDurationClass.SHORT_TERM: 0.7,
            LoadDurationClass.INSTANTANEOUS: 0.9,
        },
    },
    WoodType.GLUED_LAMINATED_TIMBER: {
        ServiceClass.SC1: {
            LoadDurationClass.PERMANENT: 0.6,
            LoadDurationClass.LONG_TERM: 0.7,
            LoadDurationClass.MEDIUM_TERM: 0.8,
            LoadDurationClass.SHORT_TERM: 0.9,
            LoadDurationClass.INSTANTANEOUS: 1.1,
        },
        ServiceClass.SC2: {
            LoadDurationClass.PERMANENT: 0.6,
            LoadDurationClass.LONG_TERM: 0.7,
            LoadDurationClass.MEDIUM_TERM: 0.8,
            LoadDurationClass.SHORT_TERM: 0.9,
            LoadDurationClass.INSTANTANEOUS: 1.1,
        },
        ServiceClass.SC3: {
            LoadDurationClass.PERMANENT: 0.5,
            LoadDurationClass.LONG_TERM: 0.55,
            LoadDurationClass.MEDIUM_TERM: 0.65,
            LoadDurationClass.SHORT_TERM: 0.7,
            LoadDurationClass.INSTANTANEOUS: 0.9,
        },
    },
    # TODO: Add k_mod values for other wood types
}


def get_modification_factor(
    wood_type: str | WoodType,
    service_class: int | ServiceClass,
    load_duration_class: str | LoadDurationClass,
) -> float:
    """
    Modification factor according to EN 1995-1-1, 3.2, table 3.1.

    :param wood_type: Material type.
    :param service_class: Service class.
    :param load_duration_class: Load duration class.
    :return: Modification factor.
    """
    wood_type = WoodType(wood_type)
    service_class = ServiceClass(service_class)
    load_duration_class = LoadDurationClass(load_duration_class)

    return MODIFICATION_FACTORS[wood_type][service_class][load_duration_class]


DEFORMATION_FACTORS = {
    WoodType.SOLID_TIMBER: {
        ServiceClass.SC1: 0.6,
        ServiceClass.SC2: 0.8,
        ServiceClass.SC3: 2.0,
    },
    WoodType.GLUED_LAMINATED_TIMBER: {
        ServiceClass.SC1: 0.6,
        ServiceClass.SC2: 0.8,
        ServiceClass.SC3: 2.0,
    },
    # TODO: Add k_def values for other wood types
}


def get_deformation_factor(
    wood_type: str | WoodType,
    service_class: int | ServiceClass,
) -> float:
    """
    Deformation factor according to EN 1995-1-1, 3.3, table 3.2.

    :param wood_type: Wood type.
    :param service_class: Service class.
    :return: Deformation factor.
    """
    wood_type = WoodType(wood_type)
    service_class = ServiceClass(service_class)

    return DEFORMATION_FACTORS[wood_type][service_class]



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

import numpy as np

from desssign.common.enums import CheckResult

if TYPE_CHECKING:
    import numpy.typing as npt
    from framesss.pre.member_1d import Member1D
    from desssign.loads.load_case_combination import DesignLoadCaseCombination


class Check:
    """Abstract class for design checks."""

    def __init__(
        self,
        code: str,
        paragraph: str,
        equation_number: str,
    ) -> None:
        """Init the Check object."""
        self.code = code
        self.paragraph = paragraph
        self.equation_number = equation_number

    @property
    def usages(self) -> npt.NDArray[np.float64]:
        """Usages of the material at every point along the member."""
        raise NotImplementedError("Method 'usage' must be implemented.")

    @property
    def max_usage(self) -> float:
        """Maximum usage of the material."""
        return float(np.max(self.usages))

    @property
    def result(self) -> CheckResult:
        """Overall design result."""
        if self.max_usage <= 1.0:
            return CheckResult(CheckResult.PASS)
        return CheckResult(CheckResult.FAIL)


class Member1DChecks:
    """
    Abstract class for performing design checks on 1D members.
    """

    def __init__(self, member: Member1D) -> None:
        self.member = member

    @abstractmethod
    def max_usage(self) -> float:
        """Maximum usage of the material."""
        raise NotImplementedError("Method 'max_usage' must be implemented.")

    @property
    def result(self) -> CheckResult:
        """Get the overall result of the design checks."""
        if self.max_usage <= 1.0:
            return CheckResult(CheckResult.PASS)
        return CheckResult(CheckResult.FAIL)

    def get_internal_forces(self, combination: DesignLoadCaseCombination) -> tuple[
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
    ]:
        """Get the internal forces for a given load case combination."""
        # Determine the default array sizes from the member dimensions
        default_length = self.member.x_local.shape[0]
        default_peak_length = self.member.results.peak_x_local.get(
            combination, np.zeros(0)
        ).shape[0]

        # Safe retrieval of internal forces using the .get method with default zero arrays
        axial = self.member.results.axial_forces.get(
            combination, np.zeros(default_length)
        )
        axial_peaks = self.member.results.peak_axial_forces.get(
            combination, np.zeros(default_peak_length)
        )

        shear_y = self.member.results.shear_forces_y.get(
            combination, np.zeros(default_length)
        )
        shear_y_peaks = self.member.results.peak_shear_forces_y.get(
            combination, np.zeros(default_peak_length)
        )

        shear_z = self.member.results.shear_forces_z.get(
            combination, np.zeros(default_length)
        )
        shear_z_peaks = self.member.results.peak_shear_forces_z.get(
            combination, np.zeros(default_peak_length)
        )

        torsion = self.member.results.torsional_moments.get(
            combination, np.zeros(default_length)
        )
        torsion_peaks = self.member.results.peak_torsional_moments.get(
            combination, np.zeros(default_peak_length)
        )

        bending_y = self.member.results.bending_moments_y.get(
            combination, np.zeros(default_length)
        )
        bending_y_peaks = self.member.results.peak_bending_moments_y.get(
            combination, np.zeros(default_peak_length)
        )

        bending_z = self.member.results.bending_moments_z.get(
            combination, np.zeros(default_length)
        )
        bending_z_peaks = self.member.results.peak_bending_moments_z.get(
            combination, np.zeros(default_peak_length)
        )

        # Concatenating the regular and peak forces/moments
        concatenated_axial = np.concatenate((axial, axial_peaks))
        concatenated_shear_y = np.concatenate((shear_y, shear_y_peaks))
        concatenated_shear_z = np.concatenate((shear_z, shear_z_peaks))
        concatenated_torsion = np.concatenate((torsion, torsion_peaks))
        concatenated_bending_y = np.concatenate((bending_y, bending_y_peaks))
        concatenated_bending_z = np.concatenate((bending_z, bending_z_peaks))

        return (
            concatenated_axial,
            concatenated_shear_y,
            concatenated_shear_z,
            concatenated_torsion,
            concatenated_bending_y,
            concatenated_bending_z,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import math

from desssign.common.design_check import Check

if TYPE_CHECKING:
    import numpy.typing as npt


class BendingCheck(Check):
    """
    Class for checking bending.

    :param m_ed: Bending moments along the member.
    :param m_rd_positive: Ultimate resistance of reinforced section.
    :param m_rd_negative: Ultimate resistance of reinforced section.
    """

    def __init__(
        self,
        m_ed: npt.NDArray[np.float64],
        m_rd_positive: float,
        m_rd_negative: float,
    ) -> None:
        """Init the BendingCheck object."""
        super().__init__("EN 1992-1-1", "", "6.1")
        self.m_ed = m_ed
        self.m_rd_positive = m_rd_positive
        self.m_rd_negative = m_rd_negative

    @property
    def usages(self) -> npt.NDArray[np.float64]:
        """Usages of the material at every point along the member."""
        usages = np.zeros(self.m_ed.shape)
        usages[self.m_ed >= 0] = self.m_ed[self.m_ed >= 0] / self.m_rd_positive
        usages[self.m_ed < 0] = self.m_ed[self.m_ed < 0] / self.m_rd_negative
        return usages


class ShearCheckWithoutShearReinforcement(Check):
    def __init__(
        self,
        v_ed: npt.NDArray[np.float64],
        f_ck: float,
        d: float,
        b_w: float,
        A_sl: float,
        c_rd_c: float = 0.18 / 1.15,
    ) -> None:
        """
        Init the ShearCheck object.

        :param f_ck: Characteristic concrete cylinder strength in MPa.
        :param d: Rebar diameter in mm.
        :param b_w: Smallest width of the cross-section in the tensile area.
        :param A_sl: Area of the tensile reinforcement.
        """
        super().__init__("EN 1992-1-1", "6.2", "6.2.a, 6.2.b")
        k = min(1 + math.sqrt(200 / d), 2.0)
        rho_l = min(A_sl / (b_w * d), 0.02)

        v_min = 0.035 * k ** (3 / 2) * f_ck ** (1 / 2)

        v_rd_c_a = (c_rd_c * k * (100 * rho_l * f_ck) ** (1 / 3)) * b_w * d
        v_rd_c_b = v_min * b_w * d

        self.v_rd_c = min(v_rd_c_a, v_rd_c_b)
        self.v_ed = v_ed

    @property
    def usages(self) -> npt.NDArray[np.float64]:
        """Usages of the material at every point along the member."""
        return np.abs(self.v_ed) / self.v_rd_c



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from desssign.common.design_check import Check

if TYPE_CHECKING:
    import numpy.typing as npt


class TensionParallelToTheGrainCheck(Check):
    """
    Class for checking tension parallel to the grain.

    EN 1995-1-1, 6.1.2, eq. (6.1)

    :param sigma_t0d: Design tensile stresses parallel to the grain along the member.
    :param f_t0d: Design tension strength parallel to the grain.
    """

    def __init__(
        self,
        sigma_t0d: npt.NDArray[np.float64],
        f_t0d: float,
    ) -> None:
        """Init the TensionParallelToTheGrain object."""
        super().__init__("EN 1995-1-1", "6.1.2", "6.1")
        self.sigma_t0d = sigma_t0d
        self.f_t0d = f_t0d

    @property
    def usages(self) -> npt.NDArray[np.float64]:
        """Usages of the material at every point along the member."""
        return np.abs(self.sigma_t0d) / self.f_t0d


class CompressionParallelToTheGrainCheck(Check):
    """
    Class for checking compression parallel to the grain.

    EN 1995-1-1, 6.1.4, 6.2

    :param sigma_c0d: Design compressive stresses parallel to the grain along the member.
    :param f_c0d: Design compressive strength parallel to the grain.
    """

    def __init__(
        self,
        sigma_c0d: npt.NDArray[np.float64],
        f_c0d: float,
    ) -> None:
        """Init the CompressionParallelToTheGrain object."""
        super().__init__("EN 1995-1-1", "6.1.4", "6.2")
        self.sigma_c0d = sigma_c0d
        self.f_c0d = f_c0d

    @property
    def usages(self) -> npt.NDArray[np.float64]:
        """Usages of the material at every point along the member."""
        return np.abs(self.sigma_c0d) / self.f_c0d


class BendingCheck(Check):
    """
    Class for checking bending.

    EN 1995-1-1, 6.1.5, eq. 6.11 & 6.12

    :param sigma_myd: Design bending stresses about the principal y-axis along the member.
    :param sigma_mzd: Design bending stresses about the principal z-axis along the member.
    :param f_myd: Design bending strength about the principal y-axis.
    :param f_mzd: Design bending strength about the principal z-axis.
    :param k_m: Factor considering re-distribution of bending stresses
                in cross-section.
    """

    def __init__(
        self,
        sigma_myd: npt.NDArray[np.float64],
        sigma_mzd: npt.NDArray[np.float64],
        f_myd: float,
        f_mzd: float,
        k_m: float = 0.7,  # TODO: 0.7 for rectangular cross-section, else 1.0
    ) -> None:
        """Init the Bending object."""
        super().__init__("EN 1995-1-1", "6.1.6", "6.11, 6.12")
        self.sigma_myd = sigma_myd
        self.sigma_mzd = sigma_mzd
        self.f_myd = f_myd
        self.f_mzd = f_mzd
        self.k_m = k_m

        self.eq_6_11 = (
            np.abs(self.sigma_myd) / self.f_myd
            + self.k_m * np.abs(self.sigma_mzd) / self.f_mzd
        )
        self.eq_6_12 = (
            self.k_m * np.abs(self.sigma_myd) / self.f_myd
            + np.abs(self.sigma_mzd) / self.f_mzd
        )

    @property
    def usages(self) -> npt.NDArray[np.float64]:
        """Usages of the material at every point along the member."""
        return np.maximum(self.eq_6_11, self.eq_6_12)


class ShearCheck(Check):
    """
    Class for checking shear stress.

    EN 1995-1-1, 6.1.7, eq. 6.13

    :param tau_d: Design shear stresses along the member.
    :param f_vd: Design shear strength.
    """

    def __init__(
        self,
        tau_d: npt.NDArray[np.float64],
        f_vd: float,
    ) -> None:
        """Init the Shear object."""
        super().__init__("EN 1995-1-1", "6.1.7", "6.13")
        self.tau_d = tau_d
        self.f_vd = f_vd

    @property
    def usages(self) -> npt.NDArray[np.float64]:
        """Usages of the material at every point along the member."""
        return np.abs(self.tau_d) / self.f_vd


class TorsionCheck(Check):
    """
    Class for checking torsion.

    EN 1995-1-1, 6.1.8, eq. 6.14

    :param t_tord: Design torsional stresses along the member.
    :param f_vd: Design shear strength.
    :param k_shape: Factor depending on the shape of the cross-section.
    """

    def __init__(
        self,
        t_tord: npt.NDArray[np.float64],
        f_vd: float,
        k_shape: float,
    ) -> None:
        """Init the Torsion object."""
        super().__init__("EN 1995-1-1", "6.1.8", "6.14")
        self.t_tord = t_tord
        self.f_vd = f_vd
        self.k_shape = k_shape

    @property
    def usages(self) -> npt.NDArray[np.float64]:
        """Usages of the material at every point along the member."""
        return np.abs(self.t_tord) / self.k_shape * self.f_vd


class CombinedBendingAndAxialTensionCheck(Check):
    """
    Class for checking combined bending and axial tension.

    EN 1995-1-1, 6.2.3, eq. 6.17

    :param sigma_t0d: Design tensile stresses parallel to the grain along the member.
    :param sigma_myd: Design bending stresses about the principal y-axis along the member.
    :param sigma_mzd: Design bending stresses about the principal z-axis along the member.
    :param f_t0d: Design tension strength parallel to the grain.
    :param f_myd: Design bending strength about the principal y-axis.
    :param f_mzd: Design bending strength about the principal z-axis.
    :param k_m: Factor considering re-distribution of bending stresses in cross-section.
    """

    def __init__(
        self,
        sigma_t0d: npt.NDArray[np.float64],
        sigma_myd: npt.NDArray[np.float64],
        sigma_mzd: npt.NDArray[np.float64],
        f_t0d: float,
        f_myd: float,
        f_mzd: float,
        k_m: float = 0.7,  # TODO: 0.7 for rectangular cross-section, else 1.0
    ) -> None:
        """Init the CombinedBendingAndAxialTension object."""
        super().__init__("EN 1995-1-1", "6.2.3", "6.17, 6.18")
        self.sigma_t0d = sigma_t0d
        self.sigma_myd = sigma_myd
        self.sigma_mzd = sigma_mzd
        self.f_t0d = f_t0d
        self.f_myd = f_myd
        self.f_mzd = f_mzd
        self.k_m = k_m

        tension = np.abs(self.sigma_t0d) / self.f_t0d
        bending_y = np.abs(self.sigma_myd) / self.f_myd
        bending_z = np.abs(self.sigma_mzd) / self.f_mzd

        self.eq_6_17: npt.NDArray[np.float64] = (
            tension + bending_y + self.k_m * bending_z
        )
        self.eq_6_18: npt.NDArray[np.float64] = (
            tension + self.k_m * bending_y + bending_z
        )

    @property
    def usages(self) -> npt.NDArray[np.float64]:
        """Usages of the material at every point along the member."""
        return np.maximum(self.eq_6_17, self.eq_6_18)


class CombinedBendingAndAxialCompressionCheck(Check):
    """
    Class for checking combined bending and axial compression.

    EN 1995-1-1, 6.2.4, eq. 6.19, 6.20

    :param sigma_c0d: Design compressive stresses parallel to the grain along the member.
    :param sigma_myd: Design bending stresses about the principal y-axis along the member.
    :param sigma_mzd: Design bending stresses about the principal z-axis along the member.
    :param f_c0d: Design compressive strength parallel to the grain.
    :param f_myd: Design bending strength about the principal y-axis.
    :param f_mzd: Design bending strength about the principal z-axis.
    :param k_m: Factor considering re-distribution of bending stresses in cross-section.
    """

    def __init__(
        self,
        sigma_c0d: npt.NDArray[np.float64],
        sigma_myd: npt.NDArray[np.float64],
        sigma_mzd: npt.NDArray[np.float64],
        f_c0d: float,
        f_myd: float,
        f_mzd: float,
        k_m: float = 0.7,  # TODO: 0.7 for rectangular cross-section, else 1.0
    ) -> None:
        """Init the CombinedBendingAndAxialCompression object."""
        super().__init__("EN 1995-1-1", "6.2.4", "6.19, 6.20")
        self.sigma_c0d = sigma_c0d
        self.sigma_myd = sigma_myd
        self.sigma_mzd = sigma_mzd
        self.f_c0d = f_c0d
        self.f_myd = f_myd
        self.f_mzd = f_mzd
        self.k_m = k_m

        compression = (np.abs(self.sigma_c0d) / self.f_c0d) ** 2
        bending_y = np.abs(self.sigma_myd) / self.f_myd
        bending_z = np.abs(self.sigma_mzd) / self.f_mzd

        self.eq_6_19: npt.NDArray[np.float64] = (
            compression + bending_y + self.k_m * bending_z
        )
        self.eq_6_20: npt.NDArray[np.float64] = (
            compression + self.k_m * bending_y + bending_z
        )

    @property
    def usages(self) -> npt.NDArray[np.float64]:
        """Usages of the material at every point along the member."""
        return np.maximum(self.eq_6_19, self.eq_6_20)


class ColumnStabilityCheck(Check):
    """
    Class for checking column stability.

    EN 1995-1-1, 6.3, eq. 6.23, 6.24

    :param sigma_c0d: Design compressive stresses parallel to the grain along the member.
    :param sigma_myd: Design bending stresses about the principal y-axis along the member.
    :param sigma_mzd: Design bending stresses about the principal z-axis along the member.
    :param f_c0d: Design compressive strength parallel to the grain.
    :param f_myd: Design bending strength about the principal y-axis.
    :param f_mzd: Design bending strength about the principal z-axis.
    :param k_cy: Instability factor for the y-axis.
    :param k_cz: Instability factor for the z-axis.
    :param k_m: Factor considering re-distribution of bending stresses in cross-section.
    """

    def __init__(
        self,
        sigma_c0d: npt.NDArray[np.float64],
        sigma_myd: npt.NDArray[np.float64],
        sigma_mzd: npt.NDArray[np.float64],
        f_c0d: float,
        f_myd: float,
        f_mzd: float,
        k_cy: float,
        k_cz: float,
        k_m: float,
    ) -> None:
        """Init the ColumnStabilityCheck object."""
        super().__init__("EN 1995-1-1", "6.3", "6.23, 6.24")
        self.sigma_c0d = sigma_c0d
        self.sigma_myd = sigma_myd
        self.sigma_mzd = sigma_mzd
        self.f_c0d = f_c0d
        self.f_myd = f_myd
        self.f_mzd = f_mzd
        self.k_cy = k_cy
        self.k_cz = k_cz
        self.k_m = k_m

        compression = (np.abs(self.sigma_c0d) / self.f_c0d) ** 2
        bending_y = np.abs(self.sigma_myd) / self.f_myd
        bending_z = np.abs(self.sigma_mzd) / self.f_mzd

        self.eq_6_23: npt.NDArray[np.float64] = (
            compression / k_cy + bending_y + self.k_m * bending_z
        )
        self.eq_6_24: npt.NDArray[np.float64] = (
            compression / k_cz + self.k_m * bending_y + bending_z
        )

    @property
    def usages(self) -> npt.NDArray[np.float64]:
        """Usages of the material at every point along the member."""
        return np.maximum(self.eq_6_23, self.eq_6_24)


class BeamStabilityCheck(Check):
    """
    Class for checking beam stability.

    EN 1995-1-1, 6.3, eq. 6.35

    :param sigma_c0d: Design compressive stresses parallel to the grain along the member.
    :param sigma_myd: Design bending stresses about the principal y-axis along the member.
    :param f_c0d: Design compressive strength parallel to the grain.
    :param f_myd: Design bending strength about the principal y-axis.
    :param k_crit: Factor which takes into account the reduced bending strength due to lateral buckling.
    :param k_cz: Instability factor for the z-axis.
    """

    def __init__(
        self,
        sigma_c0d: npt.NDArray[np.float64],
        sigma_myd: npt.NDArray[np.float64],
        f_c0d: float,
        f_myd: float,
        k_crit: float,
        k_cz: float,
    ) -> None:
        """Init the BeamStabilityCheck object."""
        super().__init__("EN 1995-1-1", "6.3", "6.35")
        self.sigma_c0d = sigma_c0d
        self.sigma_myd = sigma_myd
        self.f_c0d = f_c0d
        self.f_myd = f_myd
        self.k_crit = k_crit
        self.k_cz = k_cz

        compression = np.abs(self.sigma_c0d) / (self.k_cz * self.f_c0d)
        bending = (np.abs(self.sigma_myd) / (self.k_crit * self.f_myd)) ** 2

        self.eq_6_35: npt.NDArray[np.float64] = compression + bending

    @property
    def usages(self) -> npt.NDArray[np.float64]:
        """Usages of the material at every point along the member."""
        return self.eq_6_35



from __future__ import annotations

from typing import TYPE_CHECKING

from desssign.loads.wind.external_pressure_coefficients import (
    get_duopitch_roof_coefficient,
)
from desssign.loads.wind.external_pressure_coefficients import get_flat_roof_coefficient
from desssign.loads.wind.wind_load import WindLoad

if TYPE_CHECKING:
    from desssign.loads.wind.enums import FlatRoofType
    from desssign.loads.wind.enums import TerrainCategory
    from desssign.loads.wind.enums import WindZone


class Roof:
    """
    Abstract class for roofs.

    :ivar zone: Wind zone.
    :ivar terrain_category: Terrain category.
    :ivar b_x: Width of the building in the x-direction.
    :ivar b_y: Width of the building in the y-direction.
    """

    def __init__(
        self,
        zone: str | WindZone,
        terrain_category: str | TerrainCategory,
        b_x: float,
        b_y: float,
    ) -> None:
        """Init the Roof class."""
        self.zone = zone
        self.terrain_category = terrain_category
        self.b_x = b_x
        self.b_y = b_y


class FlatRoof(Roof):
    """
    Class for flat roofs.

    :ivar roof_type: Type of the flat roof.
    :ivar zone: Wind zone.
    :ivar terrain_category: Terrain category.
    :ivar b_x: Width of the building in the x-direction.
    :ivar b_y: Width of the building in the y-direction.
    :ivar h: Height of the building.
    :ivar h_p: Height of the parapet.
    :ivar wind_load: Instance of the :class:`WindLoad` class.
    """

    def __init__(
        self,
        roof_type: str | FlatRoofType,
        zone: str | WindZone,
        terrain_category: str | TerrainCategory,
        b_x: float,
        b_y: float,
        h: float,
        h_p: float = 0.0,
    ) -> None:
        """Init the FlatRoof class."""
        super().__init__(zone, terrain_category, b_x, b_y)
        self.roof_type = roof_type
        self.h = h
        self.h_p = h_p
        self.wind_load = WindLoad(self.zone, self.terrain_category, self.z_e)

    @property
    def z_e(self) -> float:
        """Reference height."""
        return self.h + self.h_p

    def get_roof_zones(
        self, b: float, d: float, sign: str, direction: str
    ) -> list[RoofZone]:
        """
        Get roof zones for a flat roof.

        :param b: Crosswind dimension of the building.
        :param d: Length of the building.
        :param sign: Sign of the coefficients, '+' (pressure), '-' (suction).
        :param direction: Wind direction, 'x' or 'y'.
        """
        e = min(b, 2 * self.h)
        ratio = self.h_p / self.h if self.h_p != 0.0 else None
        c_pe = {}
        for zone_label in ["F", "G", "H", "I"]:
            c_pe[zone_label] = get_flat_roof_coefficient(
                self.roof_type, zone_label, "c_pe10", sign=sign, ratio=ratio
            )

        if direction.lower() == "x":
            zones = [
                RoofZone(
                    "F",
                    0.0,
                    0.0,
                    e / 10,
                    e / 4,
                    c_pe["F"],
                    c_pe["F"] * self.wind_load.q_p,
                ),
                RoofZone(
                    "G",
                    0.0,
                    e / 4,
                    e / 10,
                    b - 2 * e / 4,
                    c_pe["G"],
                    c_pe["G"] * self.wind_load.q_p,
                ),
                RoofZone(
                    "F",
                    0.0,
                    b - e / 4,
                    e / 10,
                    e / 4,
                    c_pe["F"],
                    c_pe["F"] * self.wind_load.q_p,
                ),
            ]
            if d - e / 2 > 0:
                zones.append(
                    RoofZone(
                        "H",
                        e / 10,
                        0.0,
                        e / 2 - e / 10,
                        b,
                        c_pe["H"],
                        c_pe["H"] * self.wind_load.q_p,
                    )
                )
                zones.append(
                    RoofZone(
                        "I",
                        e / 2,
                        0.0,
                        d - e / 2,
                        b,
                        c_pe["I"],
                        c_pe["I"] * self.wind_load.q_p,
                    )
                )
            else:
                zones.append(
                    RoofZone(
                        "H",
                        e / 10,
                        0.0,
                        d - e / 10,
                        b,
                        c_pe["H"],
                        c_pe["H"] * self.wind_load.q_p,
                    )
                )

        elif direction.lower() == "y":
            zones = [
                RoofZone(
                    "F",
                    0.0,
                    0.0,
                    e / 4,
                    e / 10,
                    c_pe["F"],
                    c_pe["F"] * self.wind_load.q_p,
                ),
                RoofZone(
                    "G",
                    e / 4,
                    0.0,
                    b - 2 * e / 4,
                    e / 10,
                    c_pe["G"],
                    c_pe["G"] * self.wind_load.q_p,
                ),
                RoofZone(
                    "F",
                    b - e / 4,
                    0.0,
                    e / 4,
                    e / 10,
                    c_pe["F"],
                    c_pe["F"] * self.wind_load.q_p,
                ),
            ]

            if d - e / 2 > 0:
                zones.append(
                    RoofZone(
                        "H",
                        0.0,
                        e / 10,
                        b,
                        e / 2 - e / 10,
                        c_pe["H"],
                        c_pe["H"] * self.wind_load.q_p,
                    )
                )
                zones.append(
                    RoofZone(
                        "I",
                        0.0,
                        e / 2,
                        b,
                        d - e / 2,
                        c_pe["I"],
                        c_pe["I"] * self.wind_load.q_p,
                    )
                )
            else:
                zones.append(
                    RoofZone(
                        "H",
                        0.0,
                        e / 10,
                        b,
                        d - e / 10,
                        c_pe["H"],
                        c_pe["H"] * self.wind_load.q_p,
                    )
                )

        else:
            raise ValueError(
                f"Direction must be either 'x' or 'y', not: '{direction}'."
            )
        return zones

    @property
    def zones_x_neg_pos(self) -> list[RoofZone]:
        """Wind zones for wind in x-direction with pressure on the left and suction on the 'I' zone."""
        return self.get_roof_zones(self.b_y, self.b_x, "+", "x")

    @property
    def zones_x_neg_neg(self) -> list[RoofZone]:
        """Wind zones for wind in x-direction with pressure everywhere."""
        return self.get_roof_zones(self.b_y, self.b_x, "-", "x")

    @property
    def zones_y_neg_pos(self) -> list[RoofZone]:
        """Wind zones for wind in y-direction with pressure on the left and suction on the 'I' zone."""
        return self.get_roof_zones(self.b_x, self.b_y, "+", "y")

    @property
    def zones_y_neg_neg(self) -> list[RoofZone]:
        """Wind zones for wind in y-direction with pressure everywhere."""
        return self.get_roof_zones(self.b_x, self.b_y, "-", "y")


class MonopitchRoof(Roof):
    """
    Class for monopitched roofs.

    :ivar zone: Wind zone.
    :ivar terrain_category: Terrain category.
    :ivar b_x: Width of the building in the x-direction.
    :ivar b_y: Width of the building in the y-direction.
    :ivar h: Height of the building.
    :ivar wind_load: Instance of the :class:`WindLoad` class.
    """

    def __init__(
        self,
        zone: str | WindZone,
        terrain_category: str | TerrainCategory,
        b_x: float,
        b_y: float,
        h: float,
    ) -> None:
        """Init the MonopitchedRoof class."""
        super().__init__(zone, terrain_category, b_x, b_y)
        self.h = h
        self.wind_load = WindLoad(self.zone, self.terrain_category, self.z_e)
        raise NotImplementedError("MonopitchedRoof class is not implemented yet.")

    @property
    def z_e(self) -> float:
        """Reference height."""
        return self.h


class DuopitchRoof(Roof):
    """
    Class for duopitched roofs.

    The ridge is considered parallel to the y-axis.

    :ivar zone: Wind zone.
    :ivar terrain_category: Terrain category.
    :ivar b_x: Width of the building in the x-direction.
    :ivar b_y: Width of the building in the y-direction.
    :ivar h: Height of the building.
    :ivar pitch_angle: Pitch angle of the roof.
    """

    def __init__(
        self,
        zone: str | WindZone,
        terrain_category: str | TerrainCategory,
        b_x: float,
        b_y: float,
        h: float,
        pitch_angle: float,
    ) -> None:
        """Init the DuopitchRoof class."""
        super().__init__(zone, terrain_category, b_x, b_y)
        self.h = h
        self.pitch_angle = pitch_angle
        self.wind_load = WindLoad(self.zone, self.terrain_category, self.z_e)

    @property
    def z_e(self) -> float:
        """Reference height."""
        return self.h

    def get_roof_zones(
        self,
        b: float,
        d: float,
        signs: list[str],
        wind_direction: str,
    ) -> list[RoofZone]:
        """
        Get wind zones for a duopitch roof.

        :param b: Crosswind dimension of the building.
        :param d: Length of the building.
        :param signs: Signs of the coefficients, '+' (pressure), '-' (suction).
        :param wind_direction: Wind direction, 'x' or 'y'.
        """
        e = min(b, 2 * self.h)

        c_pe = {}

        zone_keys = ["F", "G", "H", "I"]
        zone_signs = [signs[0]] * 3 + [signs[1]] * 1
        if wind_direction == "x":
            zone_keys.append("J")
            zone_signs.append(signs[1])

        for zone_key, sign in zip(zone_keys, zone_signs):
            c_pe[zone_key] = get_duopitch_roof_coefficient(
                zone_key, "c_pe10", sign, self.pitch_angle, wind_direction
            )

        if wind_direction.lower() == "x":
            zones = [
                RoofZone(
                    "F",
                    0.0,
                    0.0,
                    e / 10,
                    e / 4,
                    c_pe["F"],
                    c_pe["F"] * self.wind_load.q_p,
                ),
                RoofZone(
                    "G",
                    0.0,
                    e / 4,
                    e / 10,
                    b - 2 * e / 4,
                    c_pe["G"],
                    c_pe["G"] * self.wind_load.q_p,
                ),
                RoofZone(
                    "F",
                    0.0,
                    b - e / 4,
                    e / 10,
                    e / 4,
                    c_pe["F"],
                    c_pe["F"] * self.wind_load.q_p,
                ),
                RoofZone(
                    "H",
                    e / 10,
                    0.0,
                    d / 2 - e / 10,
                    b,
                    c_pe["H"],
                    c_pe["H"] * self.wind_load.q_p,
                ),
                RoofZone(
                    "J",
                    d / 2,
                    0.0,
                    e / 10,
                    b,
                    c_pe["J"],
                    c_pe["J"] * self.wind_load.q_p,
                ),
                RoofZone(
                    "I",
                    d / 2 + e / 10,
                    0.0,
                    d / 2 - e / 10,
                    b,
                    c_pe["I"],
                    c_pe["I"] * self.wind_load.q_p,
                ),
            ]

        elif wind_direction.lower() == "y":
            zones = [
                RoofZone(
                    "F",
                    0.0,
                    0.0,
                    e / 4,
                    e / 10,
                    c_pe["F"],
                    c_pe["F"] * self.wind_load.q_p,
                ),
                RoofZone(
                    "G",
                    e / 4,
                    0.0,
                    b / 2 - e / 4,
                    e / 10,
                    c_pe["G"],
                    c_pe["G"] * self.wind_load.q_p,
                ),
                RoofZone(
                    "G",
                    b / 2,
                    0.0,
                    b / 2 - e / 4,
                    e / 10,
                    c_pe["G"],
                    c_pe["G"] * self.wind_load.q_p,
                ),
                RoofZone(
                    "F",
                    b - e / 4,
                    0.0,
                    e / 4,
                    e / 10,
                    c_pe["F"],
                    c_pe["F"] * self.wind_load.q_p,
                ),
                RoofZone(
                    "H",
                    0.0,
                    e / 10,
                    b / 2,
                    e / 2 - e / 10,
                    c_pe["H"],
                    c_pe["H"] * self.wind_load.q_p,
                ),
                RoofZone(
                    "H",
                    b / 2,
                    e / 10,
                    b / 2,
                    e / 2 - e / 10,
                    c_pe["H"],
                    c_pe["H"] * self.wind_load.q_p,
                ),
                RoofZone(
                    "I",
                    0.0,
                    e / 2,
                    b / 2,
                    d - e / 2,
                    c_pe["I"],
                    c_pe["I"] * self.wind_load.q_p,
                ),
                RoofZone(
                    "I",
                    b / 2,
                    e / 2,
                    b / 2,
                    d - e / 2,
                    c_pe["I"],
                    c_pe["I"] * self.wind_load.q_p,
                ),
            ]

        else:
            raise ValueError(
                f"Direction must be either 'x' or 'y', not: '{wind_direction}'."
            )

        return zones

    @property
    def zones_x_neg_neg(self) -> list[RoofZone] | None:
        """Wind zones for wind in x-direction with pressure on both sides of the roof."""
        if 45.0 < self.pitch_angle <= 75.0:
            return None

        return self.get_roof_zones(self.b_y, self.b_x, ["-", "-"], "x")

    @property
    def zones_wind_x_neg_pos(self) -> list[RoofZone] | None:
        """Wind zones for wind in x-direction with pressure on the left and suction on the right side of the roof."""
        if not (-5.0 <= self.pitch_angle <= 45.0):
            return None

        return self.get_roof_zones(self.b_y, self.b_x, ["-", "+"], "x")

    @property
    def zones_wind_x_pos_neg(self) -> list[RoofZone] | None:
        """Wind zones for wind in x-direction with suction on the left and pressure on the right side of the roof."""
        if not (5.0 <= self.pitch_angle <= 75.0):
            return None

        return self.get_roof_zones(self.b_y, self.b_x, ["+", "-"], "x")

    @property
    def zones_wind_x_pos_pos(self) -> list[RoofZone] | None:
        """Wind zones for wind in x-direction with pressure on both sides of the roof."""
        if not (5.0 <= self.pitch_angle <= 45.0):
            return None

        return self.get_roof_zones(self.b_y, self.b_x, ["+", "+"], "x")

    @property
    def zones_wind_y(self) -> list[RoofZone] | None:
        """Wind zones for wind in y-direction."""
        return self.get_roof_zones(self.b_x, self.b_y, ["-", "-"], "y")


class RoofZone:
    """
    Class representing a roof zone.

    :ivar roof_zone: Label of the roof zone.
    :ivar x_bl: X coordinate of the bottom-left corner.
    :ivar y_bl: Y coordinate of the bottom-left corner.
    :ivar b_x: Width of the zone in the x-direction.
    :ivar b_y: Width of the zone in the y-direction.
    :ivar c_pe: External pressure coefficient.
    :ivar w_e: Wind pressure.
    """

    def __init__(
        self,
        roof_zone: str,
        x_bl: float,
        y_bl: float,
        b_x: float,
        b_y: float,
        c_pe: float,
        w_e: float,
    ) -> None:
        """Init the RoofZone class."""
        self.roof_zone = roof_zone
        self.c_pe = c_pe
        self.w_e = w_e

        self.x_bl = x_bl
        self.y_bl = y_bl
        self.b_x = b_x
        self.b_y = b_y

        self.geometry = calculate_rectangle_points(x_bl, y_bl, b_x, b_y)

    def __str__(self) -> str:
        """Return a string representation of the RoofZone class."""
        return f"RoofZone({self.roof_zone}, {self.geometry}, {self.c_pe}, {self.w_e})"


def calculate_rectangle_points(
    x_bl: float, y_bl: float, dx: float, dy: float
) -> list[list[float]]:
    """
    Calculate corner points of a rectangle.

    :param x_bl: X coordinate of the bottom-left corner.
    :param y_bl: Y coordinate of the bottom-left corner.
    :param dx: Width of the rectangle.
    :param dy: Height of the rectangle.
    :return: A list of corner points of the rectangle.
    """
    return [[x_bl, y_bl], [x_bl + dx, y_bl], [x_bl + dx, y_bl + dy], [x_bl, y_bl + dy]]



from __future__ import annotations

import math
from typing import TYPE_CHECKING

from framesss.pre.member_1d import Member1D

from desssign.concrete.design_checks.member_1d_checks import ConcreteMember1DChecks

if TYPE_CHECKING:
    import numpy as np
    import numpy.typing as npt
    from framesss.enums import BeamConnection
    from framesss.enums import Element1DType
    from framesss.fea.analysis.analysis import Analysis
    from framesss.fea.node import Node

    from desssign.loads.load_case_combination import DesignLoadCaseCombination, DesignNonlinearLoadCaseCombination
    from desssign.concrete.concrete_section import ConcreteSection


class ConcreteMember1D(Member1D):
    """
    Class for 1D concrete member.

    Inherits from `Member1D` class and adds specific properties of concrete.

    :param label: A user-defined identifier for the member.
    :param element_type: Specifies the type of the element ('navier', 'timoshenko').
    :param nodes: The nodes at the ends of the member.
    :param section: The cross-section of the member.
    :param hinges: Defines the type of connections at the start and the end of the  member
                   (e.g., fixed, hinged, or semirigid) to model the rotational stiffness accurately.
    :param auxiliary_vector_xy_plane: An auxiliary vector in local xy-plane that defines the local
                                      coordinate system of the member.
    :param analysis: The :class:`Analysis` object.
    """

    section: (
        ConcreteSection  # Explicit type annotation, so that mypy can check the type
    )

    def __init__(
        self,
        label: str,
        element_type: str | Element1DType,
        nodes: list[Node],
        section: ConcreteSection,
        hinges: list[str] | list[BeamConnection],
        auxiliary_vector_xy_plane: npt.NDArray[np.float64],
        analysis: Analysis,
    ) -> None:
        """Init the WoodMember1D object."""
        super().__init__(
            label=label,
            element_type=element_type,
            nodes=nodes,
            section=section,
            hinges=hinges,
            auxiliary_vector_xy_plane=auxiliary_vector_xy_plane,
            analysis=analysis,
        )

        self.design_checks = ConcreteMember1DChecks(self)

    def perform_uls_checks(
        self, load_combinations: list[DesignLoadCaseCombination | DesignNonlinearLoadCaseCombination]
    ) -> None:
        """Perform the design checks."""
        self.design_checks.perform_uls_checks(load_combinations)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def flatten_list(mixed_list: tuple[Any] | list[Any]) -> list[Any]:
    """
    Flatten a list containing nested lists into a single list with all the elements.

    Iterates through the input list and extends the output list with the elements of any inner
    lists found. Non-list items are appended directly to the output list.

    :param mixed_list: A list potentially containing nested lists along with other items.
    :return: A single, flattened list containing all elements from the input, preserving order.
    """
    flat_list = []
    for item in mixed_list:
        if isinstance(item, (list, tuple)):  # Check if the item is a list or tuple
            flat_list.extend(flatten_list(item))  # Recursively flatten and extend
        else:
            flat_list.append(item)  # Append the item directly if it's not a list
    return flat_list



from __future__ import annotations

from framesss.pre.material import Material

from desssign.loads.enums import LoadDurationClass
from desssign.wood.constants import get_modification_factor
from desssign.wood.constants import get_partial_factor
from desssign.wood.enums import ServiceClass
from desssign.wood.enums import WoodType
from desssign.wood.strength_classes import WOOD_STRENGTH_CLASSES


class WoodMaterial(Material):
    """
    Class for structural wood material.

    Inherits from Material class and adds specific properties of wood.

    :param strength_class: Strength class of the wood material.
    :param service_class: Service class of the wood material.
    """

    def __init__(
        self,
        strength_class: str,
        service_class: int | ServiceClass,
    ) -> None:
        """Init the WoodMaterial object."""
        if strength_class not in WOOD_STRENGTH_CLASSES:
            raise ValueError(f"Invalid strength class: {strength_class}")

        self.strength_class = strength_class
        self.wood_type = (
            WoodType.SOLID_TIMBER
        )  # TODO: Add wood type to strength classes
        self.service_class = ServiceClass(service_class)

        wood_properties = WOOD_STRENGTH_CLASSES[strength_class]

        self.f_mk = wood_properties["f_mk"]

        self.f_t0k = wood_properties["f_t0k"]
        self.f_t90k = wood_properties["f_t90k"]
        self.f_c0k = wood_properties["f_c0k"]
        self.f_c90k = wood_properties["f_c90k"]
        self.f_vk = wood_properties["f_vk"]
        self.E_m0mean = wood_properties["E_m0mean"]
        self.E_m0k = wood_properties["E_m0k"]
        self.E_m90mean = wood_properties["E_m90mean"]
        self.G_mean = wood_properties["G_mean"]
        self.rho_k = wood_properties["rho_k"]
        self.rho_mean = wood_properties["rho_mean"]

        poisson = self.E_m0mean / (2 * self.G_mean) - 1

        super().__init__(
            label=strength_class,
            elastic_modulus=self.E_m0k,
            poissons_ratio=poisson,
            thermal_expansion_coefficient=0.5e-6,  # According to EN 1995-1-5, table C.1
            density=self.rho_mean,
        )

    def __repr__(self) -> str:
        """Return the string representation of the WoodMaterial object."""
        return f"WoodMaterial({self.strength_class})"

    @property
    def gamma_m(self) -> float:
        """Partial factor according to EN 1995-1-1, table 2.5."""
        return get_partial_factor(self.wood_type)

    def get_design_value(
        self,
        characteristic_value: float,
        load_duration_class: str | LoadDurationClass,
    ) -> float:
        """
        Return the design value of a characteristic value.

        EN 1995-1-1, 2.4.1, eq. (2.14)

        :param characteristic_value: Characteristic value.
        :param load_duration_class: Load duration class.
        :return: Design value.
        """
        load_duration_class = LoadDurationClass(load_duration_class)

        k_mod = get_modification_factor(
            self.wood_type, self.service_class, load_duration_class
        )
        return k_mod * characteristic_value / self.gamma_m

    @property
    def beta_c(self) -> float:
        """Factor for members within the straightness limits defined in Section 10 of EN 1995-1-1."""
        if self.wood_type == WoodType.SOLID_TIMBER:
            return 0.2
        elif self.wood_type == WoodType.GLUED_LAMINATED_TIMBER:
            return 0.1
        elif self.wood_type == WoodType.LVL:
            return 0.1
        else:
            raise AttributeError(f"Invalid wood type: {self.wood_type}")



from __future__ import annotations

import math as math

from framesss.pre.material import Material

from desssign.concrete.strength_classes import CONCRETE_STRENGTH_CLASSES
from desssign.concrete.constants import PARTIAL_FACTOR_CONCRETE, ALPHA_CC, ALPHA_CT


class ConcreteMaterial(Material):
    """
    Class for structural concrete material.

    Inherits from Material class and adds specific properties of concrete.

    :param strength_class: Strength class of the concrete material.
    """

    def __init__(
        self,
        strength_class: str | None = None,
        f_ck: float | None = None,
        gamma_c: float = PARTIAL_FACTOR_CONCRETE,
        alpha_cc: float = ALPHA_CC,
        alpha_ct: float = ALPHA_CT,
    ) -> None:
        """Initialise concrete material properties."""
        self.strength_class = strength_class

        if strength_class is not None:
            if strength_class not in CONCRETE_STRENGTH_CLASSES.keys():
                raise ValueError(
                    f"Concrete strength class '{strength_class}' not recognised. "
                    f"Choose from: {[c for c in CONCRETE_STRENGTH_CLASSES.keys()]}."
                )
            self.f_ck = CONCRETE_STRENGTH_CLASSES.get(strength_class)
        else:
            if f_ck is None:
                raise ValueError(
                    "Concrete strength class or characteristic "
                    "compressive cylinder strength must be provided."
                )
            if f_ck <= 0:
                raise ValueError(
                    "Characteristic compressive cylinder strength of concrete must be positive."
                )
            self.f_ck = f_ck

        self.gamma_c = gamma_c
        self.alpha_cc = alpha_cc
        self.alpha_ct = alpha_ct

        poisson = 0.2  # TODO: Add Poisson's ratio

        super().__init__(
            label=strength_class,
            elastic_modulus=self.e_cm,
            poissons_ratio=poisson,
            thermal_expansion_coefficient=10e-6,  # According to EN 1995-1-5, table C.1
            density=2400,
        )

    def __repr__(self) -> str:
        """Return the string representation of the WoodMaterial object."""
        return f"ConcreteMaterial({self.strength_class})"

    @property
    def f_cd(self) -> float:
        """
        Design compressive strength of concrete [Pa].

        EN 1992-1-1:2004, Eq. (3.15).
        """
        return self.alpha_cc * self.f_ck / self.gamma_c

    @property
    def f_ctd(self) -> float:
        """
        Design tensile strength of concrete [Pa].

        EN 1992-1-1:2004, Eq. (3.16).
        """
        return self.alpha_ct * self.f_ctk_0_05 / self.gamma_c

    @property
    def f_cm(self) -> float:
        """Mean value of concrete cylinder compressive strength [Pa]."""
        return self.f_ck + 8e6

    @property
    def f_ctm(self) -> float:
        """Mean value of axial tensile strength of concrete [Pa]."""
        if self.f_ck <= 50e6:
            f_ctm = 0.3 * self.f_ck ** (2 / 3)
        else:
            f_ctm = 2.12 * math.log(1 + (self.f_cm / 1e6) / 10)
        return f_ctm

    @property
    def f_ctk_0_05(self) -> float:
        """5% fractile of characteristic axial tensile strength of concrete [Pa]."""
        return 0.7 * self.f_ctm

    @property
    def f_ctk_0_95(self) -> float:
        """95% fractile of characteristic axial tensile strength of concrete [Pa]."""
        return 1.3 * self.f_ctm

    @property
    def e_cm(self) -> float:
        """Mean value of concrete compressive modulus of elasticity [Pa]."""
        return (22 * ((self.f_cm / 1e6) / 10) ** 0.3) * 1e9

    @property
    def eps_c1(self) -> float:
        """Compressive strain in the concrete at the peak stress f_cm [1]."""
        return min(0.7 * (self.f_cm / 1e6) ** 0.31, 2.8) / 1000

    @property
    def eps_cu1(self) -> float:
        """Ultimate compressive strain in the concrete [1]."""
        if self.f_ck <= 50e6:
            return 0.0035
        else:
            return (2.8 + 27 * ((98 - (self.f_cm / 1e6)) / 100) ** 4) / 1000

    @property
    def eps_c2(self) -> float:
        """Compressive strain in the concrete separating the parabolic and constant part [1]."""
        if self.f_ck <= 50e6:
            return 0.002
        else:
            return (2 + 0.085 * ((self.f_ck / 1e6) - 50) ** 0.53) / 1000

    @property
    def eps_cu2(self) -> float:
        """Ultimate compressive strain in the concrete [1]."""
        if self.f_ck <= 50e6:
            return 0.0035
        else:
            return (2.6 + 35 * ((90 - (self.f_ck / 1e6)) / 100) ** 4) / 1000

    @property
    def n(self) -> float:
        """Exponent for calculating the parabolic stress-strain curve of concrete [1]."""
        if self.f_ck <= 50e6:
            return 2.0
        else:
            return 1.4 + 23.4 * ((90 - (self.f_ck / 1e6)) / 100) ** 4

    @property
    def eps_c3(self) -> float:
        """Compressive strain in the concrete separating the linear and constant part [1]."""
        if self.f_ck <= 50e6:
            return 0.002
        else:
            return (1.75 + 0.55 * (((self.f_ck / 1e6) - 50) / 40)) / 1000

    @property
    def eps_cu3(self) -> float:
        """Ultimate compressive strain in the concrete [1]."""
        if self.f_ck <= 50e6:
            return 0.0035
        else:
            return (2.6 + 35 * ((90 - (self.f_ck / 1e6)) / 100) ** 4) / 1000

    @property
    def eta(self) -> float:
        """Parameter defining effective strength of a rectangular stress distribution [1]."""
        if self.f_ck <= 50e6:
            return 1.0
        else:
            return 1.0 - ((self.f_ck / 1e6) - 50.0) / 200.0

    @property
    def lamb(self) -> float:
        """Effective height coefficient of the compression area in rectangular stress distribution [1]."""
        if self.f_ck <= 50e6:
            return 0.8
        else:
            return 0.8 - ((self.f_ck / 1e6) - 50.0) / 200.0



from __future__ import annotations

from framesss.pre.material import Material

from desssign.concrete.strength_classes import STEEL_STRENGTH_CLASSES
from desssign.concrete.constants import PARTIAL_FACTOR_STEEL


class RebarMaterial(Material):
    """
    Class for structural steel rebar material.

    Inherits from Material class and adds specific properties of steel rebar.

    :param steel_grade: Strength class of the rebar material.
    :ivar strength_class: Steel grade according to EN 1992-1-1:2004.
    :ivar f_yk: Characteristic yield strength of steel [MPa].
    :ivar k: Coefficient for the design yield strength of steel.
    :ivar epsilon_uk: Ultimate strain of steel.
    :ivar gamma_s: Partial safety factor for steel.
    :ivar e: Young's modulus of steel [MPa].
    """

    def __init__(
        self,
        steel_grade: str | None = None,
        f_yk: float | None = None,
        k: float = 1.08,
        epsilon_uk: float = 0.050,
        gamma_s: float = PARTIAL_FACTOR_STEEL,
    ) -> None:
        """Initialise concrete material properties."""
        self.strength_class = steel_grade

        if steel_grade is not None:
            if steel_grade not in STEEL_STRENGTH_CLASSES.keys():
                raise ValueError(
                    f"Steel grade '{steel_grade}' not recognised. "
                    f"Choose from: {[c for c in STEEL_STRENGTH_CLASSES.keys()]}."
                )
            self.f_yk = STEEL_STRENGTH_CLASSES.get(steel_grade).get("f_yk")
            self.k = STEEL_STRENGTH_CLASSES.get(steel_grade).get("k")
            self.epsilon_uk = STEEL_STRENGTH_CLASSES.get(steel_grade).get("epsilon_uk")
        else:
            if f_yk is None:
                raise ValueError("Steel grade or yield strength must be provided.")
            if f_yk <= 0:
                raise ValueError("Yield strength of steel must be positive.")
            self.f_yk = f_yk
            self.k = k
            self.epsilon_uk = epsilon_uk

        self.e: float = 200_000e6
        self.gamma_s = gamma_s

        poisson = 0.3  # TODO: Add Poisson's ratio

        super().__init__(
            label=steel_grade,
            elastic_modulus=self.e,  # Convert to Pa
            poissons_ratio=poisson,
            thermal_expansion_coefficient=10e-6,
            density=7850,
        )

    def __repr__(self) -> str:
        """Return the string representation of the SteelMaterial object."""
        return f"SteelMaterial({self.strength_class})"

    @property
    def f_yd(self) -> float:
        """Design yield strength of steel [MPa]."""
        return self.f_yk / self.gamma_s

    @property
    def f_yd2(self) -> float:
        """Design strength of steel [MPa]."""
        eps_uk = self.epsilon_uk
        eps_ud = self.epsilon_ud
        f_yd = self.f_yd
        eps_sy = self.epsilon_syd
        k = self.k
        f_yd2 = f_yd * (-k * eps_sy + eps_ud + eps_uk) / (eps_sy - eps_uk)
        return f_yd2

    @property
    def f_yk2(self) -> float:
        """Characteristic strength of steel [MPa]."""
        return self.k * self.f_yk

    @property
    def epsilon_ud(self) -> float:
        """Design ultimate strain of steel."""
        return 0.9 * self.epsilon_uk

    @property
    def epsilon_syk(self) -> float:
        """Yield strain of steel."""
        return self.f_yk / self.e

    @property
    def epsilon_syd(self) -> float:
        return self.f_yd / self.e



from __future__ import annotations

from typing import TYPE_CHECKING

from framesss.enums import BeamConnection
from framesss.enums import Element1DType
from framesss.fea.analysis.analysis import Analysis
from framesss.fea.analysis.frame_xz_analysis import FrameXZAnalysis
from framesss.fea.models.model import Model

from desssign.loads.enums import LimitState
from desssign.loads.enums import LoadDurationClass
from desssign.loads.enums import LoadType
from desssign.loads.enums import SLSCombination
from desssign.loads.enums import ULSAlternativeCombination
from desssign.loads.enums import ULSCombination
from desssign.loads.enums import VariableCategory
from desssign.loads.load_case import DesignLoadCase
from desssign.loads.load_case_combination import DesignNonlinearLoadCaseCombination
from desssign.loads.load_case_combination import DesignLoadCaseCombination

from desssign.wood.wood_member import WoodMember1D
from desssign.concrete.concrete_member import ConcreteMember1D

if TYPE_CHECKING:
    from framesss.fea.node import Node

    from desssign.wood.wood_section import WoodRectangularSection
    from desssign.concrete.concrete_section import ConcreteSection


class DesignModel(Model):
    """
    Class represent the entire structural analysis model.

    Upon :class:`framesss.fea.models.Model` class, it changes
    """

    load_combinations: set[DesignLoadCaseCombination]
    nonlinear_load_combinations: set[DesignNonlinearLoadCaseCombination]
    members: set[WoodMember1D | ConcreteMember1D]

    def __init__(self, analysis: Analysis) -> None:
        """Init the DesignModel object."""
        super().__init__(analysis)

    def add_wood_member(
        self,
        label: str,
        element_type: str,
        nodes: list[Node],
        section: WoodRectangularSection,
        hinges: list[str] | tuple[str, str] = (BeamConnection.CONTINUOUS_END,) * 2,
    ) -> WoodMember1D:
        """
        Create and return new :class:`WoodMember1D` instance.

        :param label: A user-defined label for the member.
        :param element_type: Specifies the type of the element ('navier', 'timoshenko').
        :param nodes: A list of nodes at the start and the end of the member.
        :param section: The cross-section of the member.
        :param hinges: Defines the type of connections at the start and the end of the  member
                       (e.g., fixed, hinged, or semirigid) to model the rotational stiffness accurately.
        """
        elem_type = Element1DType(element_type)
        hngs = [BeamConnection(hng) for hng in hinges]

        aux = self.analysis.get_auxiliary_vector_in_local_xy_plane(nodes)
        new_member = WoodMember1D(
            label,
            elem_type,
            nodes,
            section,
            hngs,
            aux,
            self.analysis,
        )
        self.members.add(new_member)
        return new_member

    def add_concrete_member(
        self,
        label: str,
        element_type: str,
        nodes: list[Node],
        section: ConcreteSection,
        hinges: list[str] | tuple[str, str] = (BeamConnection.CONTINUOUS_END,) * 2,
    ) -> ConcreteMember1D:
        """
        Create and return new :class:`ConcreteMember1D` instance.

        :param label: A user-defined label for the member.
        :param element_type: Specifies the type of the element ('navier', 'timoshenko').
        :param nodes: A list of nodes at the start and the end of the member.
        :param section: The cross-section of the member.
        :param hinges: Defines the type of connections at the start and the end of the  member
                       (e.g., fixed, hinged, or semirigid) to model the rotational stiffness accurately.
        """
        elem_type = Element1DType(element_type)
        hngs = [BeamConnection(hng) for hng in hinges]

        aux = self.analysis.get_auxiliary_vector_in_local_xy_plane(nodes)
        new_member = ConcreteMember1D(
            label,
            elem_type,
            nodes,
            section,
            hngs,
            aux,
            self.analysis,
        )
        self.members.add(new_member)
        return new_member

    def add_design_load_case(
        self,
        label: str,
        description: str = "",
        load_type: str | LoadType = LoadType.PERMANENT,
        category: str | VariableCategory | None = None,
        load_duration_class: str | LoadDurationClass = LoadDurationClass.PERMANENT,
    ) -> DesignLoadCase:
        """
        Add and return new :class:`DesignLoadCase` instance.

        :param label: Unique user-defined label of the load case.
        :param description: A description of the load case.
        :param load_type: The type of the load case. Either 'permanent', 'variable' or 'accidental'.
        :param category: The category of the variable load case. Only required for variable load cases.
        :param load_duration_class: The load duration class of the load case.
        """
        load_type = LoadType(load_type)
        category = VariableCategory(category) if category is not None else None
        load_duration_class = (
            LoadDurationClass(load_duration_class)
            if load_duration_class is not None
            else None
        )
        if load_type == LoadType.VARIABLE and category is None:
            raise ValueError("Variable load cases require a category.")
        if load_duration_class is None:
            raise ValueError("Load duration class must be provided.")

        new_case = DesignLoadCase(
            label, load_type, category, load_duration_class, description
        )
        self.load_cases.add(new_case)
        return new_case

    def add_design_load_case_combination(
        self,
        label: str,
        limit_state: str | LimitState,
        combination_type: str | SLSCombination | ULSCombination,
        permanent_cases: list[DesignLoadCase],
        leading_variable_case: DesignLoadCase | None,
        other_variable_cases: list[DesignLoadCase],
        is_nonlinear: bool = False,
    ) -> (
        DesignLoadCaseCombination
        | DesignNonlinearLoadCaseCombination
        | tuple[DesignLoadCaseCombination, DesignLoadCaseCombination]
        | tuple[DesignNonlinearLoadCaseCombination, DesignNonlinearLoadCaseCombination]
    ):
        """
        Add and return new :class:`LoadCaseCombination` instance.

        :param label: The label of the load case combination.
        :param limit_state: The limit state of the combination group. Either 'ULS' or 'SLS'.
        :param combination_type: The type of the combination group. For ULS: basic, alternative or accidental,
                                 for SLS: characteristic, frequent or quasi-permanent.
        :param permanent_cases: A list of permanent load cases.
        :param leading_variable_case: The leading variable load case.
        :param other_variable_cases: A list of other variable load cases.
        :param is_nonlinear: Flag if the combination is nonlinear.
        """
        CombinationClass = (
            DesignNonlinearLoadCaseCombination
            if is_nonlinear
            else DesignLoadCaseCombination
        )
        if combination_type == ULSCombination.ALTERNATIVE:
            new_combination_a = CombinationClass(
                label=f"{label}(a)",
                limit_state=limit_state,
                combination_type=combination_type,
                permanent_cases=permanent_cases,
                leading_variable_case=leading_variable_case,
                other_variable_cases=other_variable_cases,
                alternative_combination=ULSAlternativeCombination.REDUCED_VARIABLE,
            )

            new_combination_b = CombinationClass(
                label=f"{label}(b)",
                limit_state=limit_state,
                combination_type=combination_type,
                permanent_cases=permanent_cases,
                leading_variable_case=leading_variable_case,
                other_variable_cases=other_variable_cases,
                alternative_combination=ULSAlternativeCombination.REDUCED_PERMANENT,
            )

            if is_nonlinear:
                self.nonlinear_load_combinations.add(new_combination_a)
                self.nonlinear_load_combinations.add(new_combination_b)
            else:
                self.load_combinations.add(new_combination_a)
                self.load_combinations.add(new_combination_b)
            return new_combination_a, new_combination_b

        new_combination = CombinationClass(
            label=label,
            limit_state=limit_state,
            combination_type=combination_type,
            permanent_cases=permanent_cases,
            leading_variable_case=leading_variable_case,
            other_variable_cases=other_variable_cases,
        )
        if is_nonlinear:
            self.nonlinear_load_combinations.add(new_combination)
        else:
            self.load_combinations.add(new_combination)
        return new_combination

    def perform_uls_checks(self) -> None:
        """Perform ULS checks on the model members."""
        combinations = [
            comb
            for comb in self.load_combinations.union(self.nonlinear_load_combinations)
            if comb.limit_state == LimitState.ULS
        ]

        for member in self.members:
            member.perform_uls_checks(combinations)


class DesignModelFrameXZ(DesignModel):
    """
    Class represent the entire structural analysis model.
    """

    def __init__(self) -> None:
        """Init the DesignModelFrameXZ object."""
        super().__init__(FrameXZAnalysis())



from __future__ import annotations

from framesss.pre.cases import LoadCase

from desssign.loads.enums import LoadDurationClass
from desssign.loads.enums import LoadType
from desssign.loads.enums import VariableCategory


class DesignLoadCase(LoadCase):
    """
    Represent a load case in the design process.

    :param label: The label of the load case.
    :param load_type: The type of the load case. Either permanent, variable or accidental.
    :param category: The category of the variable load case. Only required for variable load cases.
    :param load_duration_class: The load duration class of the load case.
    :param description: A description of the load case.
    """

    def __init__(
        self,
        label: str,
        load_type: str | LoadType,
        category: str | VariableCategory | None = None,
        load_duration_class: str | LoadDurationClass = LoadDurationClass.PERMANENT,
        description: str = "",
    ) -> None:
        """Init the DesignLoadCase class."""
        super().__init__(label)
        self.load_type = LoadType(load_type)

        if self.load_type == LoadType.VARIABLE:
            self.category: VariableCategory | None = VariableCategory(category)
        else:
            self.category = None

        self.load_duration_class = LoadDurationClass(load_duration_class)

        self.description = description

    def __repr__(self) -> str:
        """Return a string representation of the DesignLoadCase object."""
        return (
            f"{self.__class__.__name__}("
            f"label={self.label}, "
            f"load_type={self.load_type}, "
            f"category={self.category})"
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from framesss.pre.section import PolygonalSection

if TYPE_CHECKING:
    from desssign.concrete.concrete_material import ConcreteMaterial
    from desssign.concrete.rebar_material import RebarMaterial


class ConcreteSection(PolygonalSection):
    """
    Class for concrete section.

    Inherits from `Section` class and adds specific properties of concrete section.

    :param label: Label of the section.
    :param points: Points defining boundary of the section.
    :param material: Material of the section.
    :param b_w: Smallest width of the cross-section in the tensile area.
    :param A_sl: Area of the tensile reinforcement.
    :param m_rd_positive: Bending moment resistance.
    :param m_rd_negative: Bending moment resistance.
    """

    material: (
        ConcreteMaterial  # Explicit type annotation, so that mypy can check the type
    )

    def __init__(
        self,
        label: str,
        points: list[list[float]],
        material: ConcreteMaterial,
        b_w: float,
        A_sl: float,
        m_rd_positive: float,
        m_rd_negative: float,
    ) -> None:
        super().__init__(label=label, points=points, material=material)

        self.b_w = b_w
        self.A_sl = A_sl
        self.m_rd_positive = m_rd_positive
        self.m_rd_negative = m_rd_negative



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from desssign.common.design_check import Member1DChecks
from desssign.wood.design_checks.design_check import BeamStabilityCheck
from desssign.wood.design_checks.design_check import ColumnStabilityCheck
from desssign.wood.design_checks.design_check import (
    CombinedBendingAndAxialCompressionCheck,
)
from desssign.wood.design_checks.design_check import CombinedBendingAndAxialTensionCheck
from desssign.wood.design_checks.design_check import ShearCheck

if TYPE_CHECKING:
    from desssign.loads.load_case_combination import DesignLoadCaseCombination
    from desssign.wood.wood_member import WoodMember1D


class WoodMember1DChecks(Member1DChecks):
    """
    Class for performing design checks on 1D members.

    :param member: The 1D wood member.
    """

    member: WoodMember1D  # Explicit type annotation, so that mypy can check the type

    def __init__(self, member: WoodMember1D) -> None:
        """Init the WoodMember1DChecks object."""
        super().__init__(member=member)

        self.column_stability: dict[DesignLoadCaseCombination, ColumnStabilityCheck] = (
            {}
        )
        self.beam_stability: dict[DesignLoadCaseCombination, BeamStabilityCheck] = {}

        self.shear_check: dict[DesignLoadCaseCombination, ShearCheck] = {}
        self.tension_with_bending_check: dict[
            DesignLoadCaseCombination, CombinedBendingAndAxialTensionCheck
        ] = {}
        self.compression_with_bending_check: dict[
            DesignLoadCaseCombination, CombinedBendingAndAxialCompressionCheck
        ] = {}

    @property
    def max_usage(self) -> float:
        """Get the maximum usage of the design checks."""
        max_usages = [
            check.max_usage
            for check in (
                *self.column_stability.values(),
                *self.beam_stability.values(),
                *self.shear_check.values(),
                *self.tension_with_bending_check.values(),
                *self.compression_with_bending_check.values(),
            )
        ]
        return max(max_usages)

    def perform_uls_checks(
        self, load_case_combinations: list[DesignLoadCaseCombination]
    ) -> None:
        """
        Perform all design checks on the member for given load case combinations.

        :param load_case_combinations: The list of load case combinations to check.
        """
        self.perform_column_stability_checks(load_case_combinations)
        self.perform_beam_stability_checks(load_case_combinations)
        self.perform_shear_checks(load_case_combinations)
        self.perform_tension_with_bending_checks(load_case_combinations)
        self.perform_compression_with_bending_checks(load_case_combinations)

    def perform_column_stability_checks(
        self,
        load_case_combinations: list[DesignLoadCaseCombination],
    ) -> None:
        """Perform the column stability checks for the given load case combinations."""
        for combination in load_case_combinations:
            axial, _, _, _, bending_y, bending_z = self.get_internal_forces(combination)

            sigma_c0d = axial / self.member.section.area_x
            sigma_c0d[sigma_c0d >= 0] = 0

            sigma_myd = bending_y / self.member.section.W_y
            sigma_mzd = bending_z / self.member.section.W_z

            f_c0d = self.member.section.material.get_design_value(
                characteristic_value=self.member.section.material.f_c0k,
                load_duration_class=combination.load_duration_class,
            )

            f_md = self.member.section.material.get_design_value(
                characteristic_value=self.member.section.material.f_mk,
                load_duration_class=combination.load_duration_class,
            )

            self.column_stability[combination] = ColumnStabilityCheck(
                sigma_c0d=sigma_c0d,
                sigma_myd=sigma_myd,
                sigma_mzd=sigma_mzd,
                f_c0d=f_c0d,
                f_myd=f_md,
                f_mzd=f_md,
                k_cy=self.member.k_cy,
                k_cz=self.member.k_cz,
                k_m=self.member.section.k_m,
            )

    def perform_beam_stability_checks(
        self,
        load_case_combinations: list[DesignLoadCaseCombination],
    ) -> None:
        """Perform the beam stability checks for the given load case combinations."""
        for combination in load_case_combinations:
            axial, _, _, _, bending_y, bending_z = self.get_internal_forces(combination)

            sigma_c0d = axial / self.member.section.area_x
            sigma_c0d[sigma_c0d >= 0] = 0

            sigma_myd = bending_y / self.member.section.W_y

            f_c0d = self.member.section.material.get_design_value(
                characteristic_value=self.member.section.material.f_c0k,
                load_duration_class=combination.load_duration_class,
            )

            f_md = self.member.section.material.get_design_value(
                characteristic_value=self.member.section.material.f_mk,
                load_duration_class=combination.load_duration_class,
            )

            self.beam_stability[combination] = BeamStabilityCheck(
                sigma_c0d=sigma_c0d,
                sigma_myd=sigma_myd,
                f_c0d=f_c0d,
                f_myd=f_md,
                k_crit=self.member.k_crit,
                k_cz=self.member.k_cz,
            )

    def perform_shear_checks(
        self,
        load_case_combinations: list[DesignLoadCaseCombination],
    ) -> None:
        """Perform the shear checks for the given load case combinations."""
        for combination in load_case_combinations:
            _, shear_y, shear_z, _, _, _ = self.get_internal_forces(combination)

            k_cr = self.member.section.k_cr
            tau_d = 3 * np.abs(shear_z) / (2 * k_cr * self.member.section.area_z)

            f_vd = self.member.section.material.get_design_value(
                characteristic_value=self.member.section.material.f_vk,
                load_duration_class=combination.load_duration_class,
            )

            self.shear_check[combination] = ShearCheck(
                tau_d=tau_d,
                f_vd=f_vd,
            )

    def perform_tension_with_bending_checks(
        self,
        load_case_combinations: list[DesignLoadCaseCombination],
    ) -> None:
        """Perform the tension with bending checks for the given load case combinations."""
        for combination in load_case_combinations:
            axial, _, _, _, bending_y, bending_z = self.get_internal_forces(combination)

            sigma_t0d = axial / self.member.section.area_x
            sigma_t0d[sigma_t0d < 0] = 0

            sigma_myd = bending_y / self.member.section.W_y
            sigma_mzd = bending_z / self.member.section.W_z

            f_cd = self.member.section.material.get_design_value(
                characteristic_value=self.member.section.material.f_c0k,
                load_duration_class=combination.load_duration_class,
            )
            f_md = self.member.section.material.get_design_value(
                characteristic_value=self.member.section.material.f_mk,
                load_duration_class=combination.load_duration_class,
            )

            self.tension_with_bending_check[combination] = (
                CombinedBendingAndAxialTensionCheck(
                    sigma_t0d=sigma_t0d,
                    sigma_myd=sigma_myd,
                    sigma_mzd=sigma_mzd,
                    f_t0d=f_cd,
                    f_myd=f_md,
                    f_mzd=f_md,
                    k_m=self.member.section.k_m,
                )
            )

    def perform_compression_with_bending_checks(
        self,
        load_case_combinations: list[DesignLoadCaseCombination],
    ) -> None:
        """Perform the compression with bending checks for the given load case combinations."""
        for combination in load_case_combinations:
            axial, _, _, _, bending_y, bending_z = self.get_internal_forces(combination)

            sigma_c0d = axial / self.member.section.area_x
            sigma_c0d[sigma_c0d >= 0] = 0

            sigma_myd = bending_y / self.member.section.W_y
            sigma_mzd = bending_z / self.member.section.W_z

            f_c0d = self.member.section.material.get_design_value(
                characteristic_value=self.member.section.material.f_c0k,
                load_duration_class=combination.load_duration_class,
            )
            f_md = self.member.section.material.get_design_value(
                characteristic_value=self.member.section.material.f_mk,
                load_duration_class=combination.load_duration_class,
            )

            self.compression_with_bending_check[combination] = (
                CombinedBendingAndAxialCompressionCheck(
                    sigma_c0d=sigma_c0d,
                    sigma_myd=sigma_myd,
                    sigma_mzd=sigma_mzd,
                    f_c0d=f_c0d,
                    f_myd=f_md,
                    f_mzd=f_md,
                    k_m=self.member.section.k_m,
                )
            )



from __future__ import annotations

from itertools import product
from typing import TYPE_CHECKING

from desssign.loads.enums import LimitState
from desssign.loads.enums import LoadType
from desssign.loads.enums import SLSCombination
from desssign.loads.enums import ULSAlternativeCombination
from desssign.loads.enums import ULSCombination
from desssign.loads.load_case_combination import (
    DesignLoadCaseCombination,
    DesignNonlinearLoadCaseCombination,
)
from desssign.utils import flatten_list

if TYPE_CHECKING:
    from desssign.loads.load_case_group import DesignLoadCaseGroup


class CombinationsGenerator:
    """
    A class used to generate all possible combinations of load cases for a specific limit state.

    :ivar limit_state: The limit state of the combination group. Either ULS or SLS.
    :ivar combination_type: The type of the combination group. For ULS: basic, alternative or accidental,
                            for SLS: characteristic, frequent or quasi-permanent.
    """

    def __init__(
        self,
        limit_state: str | LimitState,
        combination_type: str | SLSCombination | ULSCombination,
    ) -> None:
        """
        Initialize the CombinationsGenerator class.

        :param limit_state: The limit state of the combination group. Either ULS or SLS.
        :type limit_state: str | LimitState
        :param combination_type: The type of the combination group. For ULS: basic, alternative or accidental,
                                 for SLS: characteristic, frequent or quasi-permanent.
        :type combination_type: str | SLSCombination | ULSCombination
        :raises AttributeError: If the combination type does not match the limit state.
        """
        self.limit_state = LimitState(limit_state)

        if self.limit_state == LimitState.ULS:
            self.combination_type = ULSCombination(combination_type)
        elif self.limit_state == LimitState.SLS:
            self.combination_type = SLSCombination(combination_type)
        else:
            raise AttributeError(
                f"Can't set combination type: '{combination_type}' to limit state: '{limit_state}'."
            )

    def generate_combinations(
        self,
        *args: list[DesignLoadCaseGroup,],
        start_numbering_from: int = 1,
        is_nonlinear: bool = False,
    ) -> list[DesignLoadCaseCombination] | list[DesignNonlinearLoadCaseCombination]:
        """
        Generate all possible combinations of load cases.

        :param start_numbering_from: The number to start the combination numbering from.
        :param args: Variable length argument list of LoadCaseGroup lists.
        :param is_nonlinear: Flag to indicate if the combination is for nonlinear analysis.
        return:A list of all generated combinations of load cases.
        """
        generated_combinations = []
        # get all possible combinations
        all_iterables = []
        for load_groups in args:
            all_iterables.append(
                [load_group.combinations for load_group in load_groups]
            )

        combinations = []
        for iterables in all_iterables:
            combinations.extend(
                [flatten_list(combination) for combination in product(*iterables)]
            )

        # get all possible unique combinations
        unique_combinations = []
        for combination in combinations:
            if combination not in unique_combinations:
                unique_combinations.append(combination)

        label = "CO"
        description = f"{self.limit_state.value.upper()}-{self.combination_type.value}"

        c = start_numbering_from

        CombinationClass = (
            DesignNonlinearLoadCaseCombination
            if is_nonlinear
            else DesignLoadCaseCombination
        )

        for unique_combination in unique_combinations:
            permanent_cases = [
                case
                for case in unique_combination
                if case.load_type == LoadType.PERMANENT
            ]
            variable_cases = [
                case
                for case in unique_combination
                if case.load_type == LoadType.VARIABLE
            ]

            if variable_cases:
                for i, leading_variable_case in enumerate(variable_cases):
                    # loop through every possible combination of leading + other variable for this unique combination
                    other_variable_cases = variable_cases[:i] + variable_cases[i + 1 :]

                    if self.combination_type == ULSCombination.ALTERNATIVE:
                        generated_combinations.append(
                            CombinationClass(
                                label=f"{label}{c}a",
                                description=description,
                                limit_state=self.limit_state,
                                combination_type=self.combination_type,
                                permanent_cases=permanent_cases,
                                leading_variable_case=leading_variable_case,
                                other_variable_cases=other_variable_cases,
                                alternative_combination=ULSAlternativeCombination.REDUCED_VARIABLE,
                            )
                        )
                        generated_combinations.append(
                            CombinationClass(
                                label=f"{label}{c}b",
                                description=description,
                                limit_state=self.limit_state,
                                combination_type=self.combination_type,
                                permanent_cases=permanent_cases,
                                leading_variable_case=leading_variable_case,
                                other_variable_cases=other_variable_cases,
                                alternative_combination=ULSAlternativeCombination.REDUCED_PERMANENT,
                            )
                        )
                    else:
                        generated_combinations.append(
                            CombinationClass(
                                label=f"{label}{c}",
                                description=description,
                                limit_state=self.limit_state,
                                combination_type=self.combination_type,
                                permanent_cases=permanent_cases,
                                leading_variable_case=leading_variable_case,
                                other_variable_cases=other_variable_cases,
                            )
                        )

                    c += 1
            else:  # in case there are only permanent cases
                if self.combination_type == ULSCombination.ALTERNATIVE:
                    generated_combinations.append(
                        CombinationClass(
                            label=f"{label}{c}a",
                            description=description,
                            limit_state=self.limit_state,
                            combination_type=self.combination_type,
                            permanent_cases=permanent_cases,
                            leading_variable_case=None,
                            other_variable_cases=[],
                            alternative_combination=ULSAlternativeCombination.REDUCED_VARIABLE,
                        )
                    )
                    generated_combinations.append(
                        CombinationClass(
                            label=f"{label}{c}b",
                            description=description,
                            limit_state=self.limit_state,
                            combination_type=self.combination_type,
                            permanent_cases=permanent_cases,
                            leading_variable_case=None,
                            other_variable_cases=[],
                            alternative_combination=ULSAlternativeCombination.REDUCED_PERMANENT,
                        )
                    )
                else:
                    generated_combinations.append(
                        CombinationClass(
                            label=f"{label}{c}",
                            description=description,
                            limit_state=self.limit_state,
                            combination_type=self.combination_type,
                            permanent_cases=permanent_cases,
                            leading_variable_case=None,
                            other_variable_cases=[],
                        )
                    )
                c += 1

        return generated_combinations



from framesss.enums import CaseInsensitiveStrEnum


class LoadBehavior(CaseInsensitiveStrEnum):
    """Enumeration of possible behavior of action."""

    UNFAVOURABLE = "unfavourable"
    FAVOURABLE = "favourable"


class LoadType(CaseInsensitiveStrEnum):
    """Enumeration of possible type of action."""

    PERMANENT = "permanent"
    VARIABLE = "variable"
    ACCIDENTAL = "accidental"


class LimitState(CaseInsensitiveStrEnum):
    """
    Enumeration of limit states.

    :cvar ULS: Ultimate limit state.
    :cvar SLS: Serviceability limit state.
    """

    ULS = "uls"
    SLS = "sls"


class UltimateLimitStates(CaseInsensitiveStrEnum):
    """Enumeration of ultimate limit states."""

    EQU = "equ"
    STR = "str"
    GEO = "geo"
    FAT = "fat"
    UPL = "upl"
    HYD = "hyd"


class ULSCombination(CaseInsensitiveStrEnum):
    """
    Enumeration of ultimate limit state combinations according to EN 1990.

    :cvar BASIC: Basic combination according to equation 6.10 of EN 1990.
    :cvar ALTERNATIVE: Alternative combinations according to equations 6.10a and 6.10b of EN 1990.
                       In this case, two variants of the combinations are considered in calculations,
                       one with reduced constant load cases and the other with reduced principal
                       variable load cases.
    :cvar ACCIDENTAL: Accidental combinations according to equations 6.11 of EN 1990.
    """

    BASIC = "basic"
    ALTERNATIVE = "alternative"
    ACCIDENTAL = "accidental"


class ULSAlternativeCombination(CaseInsensitiveStrEnum):
    """
    Enumeration of ULS alternative combination types.

    :cvar REDUCED_PERMANENT: Combination with reduced permanent load cases.
    :cvar REDUCED_VARIABLE: Combination with reduced leading variable load case.
    """

    REDUCED_VARIABLE = "6.10a"
    REDUCED_PERMANENT = "6.10b"


class SLSCombination(CaseInsensitiveStrEnum):
    """
    Enumeration of serviceability limit state combinations according to EN 1990.

    :cvar CHARACTERISTIC: Combination according to equations 6.14 of EN 1990.
    :cvar FREQUENT: Combination according to equation 6.15 of EN 1990.
    :cvar QUASIPERMANENT: Combination according to equation 6.16 of EN 1990.
    """

    CHARACTERISTIC = "characteristic"
    FREQUENT = "frequent"
    QUASIPERMANENT = "quasipermanent"


class VariableCategory(CaseInsensitiveStrEnum):
    """Enumeration of possible variable loads."""

    A = "a"
    B = "b"
    C = "c"
    D = "d"
    E = "e"
    F = "f"
    G = "g"
    H = "h"
    SNOW_ABOVE_1000_M = "snow > 1000 m"
    SNOW_BELLOW_1000_M = "snow < 1000 m"
    WIND = "wind"
    TEMPERATURE = "temperature"


class LoadCaseRelation(CaseInsensitiveStrEnum):
    """
    Enumeration of possible relation of load cases.

    :cvar EXCLUSIVE: Load cases from the same load group will never act together.
    :cvar STANDARD: Load cases from the same load group may (or may not) act together.
    :cvar TOGETHER: Load cases from the same load group always act together.
    """

    EXCLUSIVE = "exclusive"
    STANDARD = "standard"
    TOGETHER = "together"


class LoadDurationClass(CaseInsensitiveStrEnum):
    """
    Enum for load duration classes.

    EN 1995-1-1, 2.3.1.2 Load-duration classes.

    The load-duration classes are characterised by the effect of a constant load acting for a
    certain period of time in the life of the structure. For a variable action the appropriate class shall
    be determined on the basis of an estimate of the typical variation of the load with time.

    :cvar PERMANENT: more than 10 years, e.g. self-weight
    :cvar LONG_TERM: 6 months - 10 years, e.g. storage
    :cvar MEDIUM_TERM: 1 week - 6 months, e.g. imposed floor load, snow
    :cvar SHORT_TERM: less than 1 week, e.g. snow, wind
    :cvar INSTANTANEOUS: instantaneous, e.g. wind, accidental load
    """

    PERMANENT = "permanent"
    LONG_TERM = "long-term"
    MEDIUM_TERM = "medium-term"
    SHORT_TERM = "short-term"
    INSTANTANEOUS = "instantaneous"


LOAD_DURATION_MAPPING = {
    LoadDurationClass.INSTANTANEOUS: 1,
    LoadDurationClass.SHORT_TERM: 2,
    LoadDurationClass.MEDIUM_TERM: 3,
    LoadDurationClass.LONG_TERM: 4,
    LoadDurationClass.PERMANENT: 5,
}



from desssign.loads.enums import LoadBehavior
from desssign.loads.enums import LoadType
from desssign.loads.enums import VariableCategory

PSI_FACTORS = {
    VariableCategory.A: {
        "psi_0": 0.7,
        "psi_1": 0.5,
        "psi_2": 0.3,
    },
    VariableCategory.B: {
        "psi_0": 0.7,
        "psi_1": 0.5,
        "psi_2": 0.3,
    },
    VariableCategory.C: {
        "psi_0": 0.7,
        "psi_1": 0.7,
        "psi_2": 0.6,
    },
    VariableCategory.D: {
        "psi_0": 0.7,
        "psi_1": 0.7,
        "psi_2": 0.6,
    },
    VariableCategory.E: {
        "psi_0": 1.0,
        "psi_1": 0.9,
        "psi_2": 0.8,
    },
    VariableCategory.F: {
        "psi_0": 0.7,
        "psi_1": 0.7,
        "psi_2": 0.6,
    },
    VariableCategory.G: {
        "psi_0": 0.7,
        "psi_1": 0.5,
        "psi_2": 0.3,
    },
    VariableCategory.H: {
        "psi_0": 0.0,
        "psi_1": 0.0,
        "psi_2": 0.0,
    },
    VariableCategory.SNOW_ABOVE_1000_M: {
        "psi_0": 0.7,
        "psi_1": 0.5,
        "psi_2": 0.2,
    },
    VariableCategory.SNOW_BELLOW_1000_M: {
        "psi_0": 0.5,
        "psi_1": 0.2,
        "psi_2": 0.0,
    },
    VariableCategory.WIND: {
        "psi_0": 0.6,
        "psi_1": 0.2,
        "psi_2": 0.0,
    },
    VariableCategory.TEMPERATURE: {
        "psi_0": 0.6,
        "psi_1": 0.5,
        "psi_2": 0.0,
    },
}


GAMMA_VALUES = {
    "Set B": {
        LoadType.PERMANENT: {
            LoadBehavior.FAVOURABLE: 1.0,
            LoadBehavior.UNFAVOURABLE: 1.35,
        },
        LoadType.VARIABLE: {
            LoadBehavior.FAVOURABLE: 0.0,
            LoadBehavior.UNFAVOURABLE: 1.5,
        },
    },
    "Set C": {
        LoadType.PERMANENT: {
            LoadBehavior.FAVOURABLE: 1.0,
            LoadBehavior.UNFAVOURABLE: 1.0,
        },
        LoadType.VARIABLE: {
            LoadBehavior.FAVOURABLE: 0.0,
            LoadBehavior.UNFAVOURABLE: 1.3,
        },
    },
}

XI = 0.85



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from desssign.loads.snow.enums import SnowZone
    from desssign.loads.snow.enums import Topography

from desssign.loads.snow.constants import EXPOSURE_COEFFICIENTS
from desssign.loads.snow.constants import SNOW_LOAD_ON_THE_GROUND


def calculate_shape_coefficient(
    roof_pitch_angle: float, is_prevented_sliding: bool = False
) -> float:
    """
    Calculate the shape coefficient for snow load calculations.

    This function takes in account only monopitched and pitched roofs.

    :param roof_pitch_angle: The angle of pitch of the roof in degrees.
    :param is_prevented_sliding: Whether the snow is prevented from sliding off the roof.
                                 Where snow fences or other obstructions exist or where
                                 the lower edge of the roof is terminated with a parapet.
    :return: The shape coefficients.
    """
    if is_prevented_sliding:
        return 0.8
    else:
        if 0 <= roof_pitch_angle <= 30:
            return 0.8
        elif 30 < roof_pitch_angle < 60:
            return 0.8 * (60 - roof_pitch_angle) / 30
        else:
            return 0.0


def calculate_snow_load_on_the_roof(
    roof_pitch_angle: float,
    snow_zone: str | SnowZone,
    topography: str | Topography,
    thermal_coefficient: float = 1.0,
    is_prevented_sliding: bool = False,
) -> float:
    """
    Calculate the snow load on the roof according to EN 1991-1-3.

    :param roof_pitch_angle: The angle of pitch of the roof in degrees.
    :param snow_zone: The snow zone according to CZ NA.
    :param topography: The topography considering future development around the site.
    :param thermal_coefficient: The thermal coefficient.
    :param is_prevented_sliding: Whether the snow is prevented from sliding off the roof.
                                 Where snow fences or other obstructions exist or where
                                 the lower edge of the roof is terminated with a parapet.
    """
    mu_1 = calculate_shape_coefficient(roof_pitch_angle, is_prevented_sliding)
    c_e = EXPOSURE_COEFFICIENTS[topography]
    c_t = thermal_coefficient
    s_k = SNOW_LOAD_ON_THE_GROUND[snow_zone]

    return mu_1 * c_e * c_t * s_k



from __future__ import annotations

from desssign.loads.enums import LoadBehavior
from desssign.loads.enums import SLSCombination
from desssign.loads.enums import ULSAlternativeCombination
from desssign.loads.enums import ULSCombination
from desssign.loads.enums import VariableCategory
from desssign.loads.load_case import DesignLoadCase
from desssign.loads.load_combination_generator.constants import GAMMA_VALUES
from desssign.loads.load_combination_generator.constants import PSI_FACTORS
from desssign.loads.load_combination_generator.constants import XI


def generate_combination(
    permanent_cases: list[DesignLoadCase],
    leading_variable_case: DesignLoadCase | None,
    other_variable_cases: list[DesignLoadCase],
    combination: SLSCombination | ULSCombination,
    alternative_combination: ULSAlternativeCombination | None = None,
) -> tuple[dict[DesignLoadCase, float], str]:
    """
    Generate a combination of load cases according to the limit state and combination type.

    :param permanent_cases: A list of permanent load cases.
    :param leading_variable_case: The leading variable load case.
    :param other_variable_cases: A list of other variable load cases.
    :param combination: Type of SLS or ULS combination.
    :param alternative_combination: Specifier for the used equation, only required for ULS alternative combinations.
                                    Either '6.10a' or '6.10b'.
    :raises AttributeError: If the combination type is unknown.
    """
    if combination == SLSCombination.CHARACTERISTIC:
        return generate_sls_characteristic_combination(
            permanent_cases, leading_variable_case, other_variable_cases
        )

    if combination == SLSCombination.FREQUENT:
        return generate_sls_frequent_combination(
            permanent_cases, leading_variable_case, other_variable_cases
        )

    if combination == SLSCombination.QUASIPERMANENT:
        return generate_sls_quasipermanent_combination(
            permanent_cases, leading_variable_case, other_variable_cases
        )

    if combination == ULSCombination.BASIC:
        return generate_uls_basic_combination(
            permanent_cases, leading_variable_case, other_variable_cases
        )

    if combination == ULSCombination.ALTERNATIVE:
        if alternative_combination == ULSAlternativeCombination.REDUCED_VARIABLE:
            return generate_uls_alternative_a_combination(
                permanent_cases, leading_variable_case, other_variable_cases
            )

        if alternative_combination == ULSAlternativeCombination.REDUCED_PERMANENT:
            return generate_uls_alternative_b_combination(
                permanent_cases, leading_variable_case, other_variable_cases
            )

        raise AttributeError(
            f"Unknown alternative combination: '{alternative_combination}'."
        )

    raise AttributeError(f"Unknown combination: '{combination}'.")


def generate_sls_characteristic_combination(
    permanent_cases: list[DesignLoadCase],
    leading_variable_case: DesignLoadCase | None,
    other_variable_cases: list[DesignLoadCase],
) -> tuple[dict[DesignLoadCase, float], str]:
    """
    Generate a characteristic combination of load cases for serviceability limit state.

    Combination is generated according to equations 6.14 of EN 1990.
    The combination is normally used for irreversible limit states.

    :param permanent_cases: A list of permanent load cases.
    :param leading_variable_case: The leading variable load case.
    :param other_variable_cases: A list of other variable load cases.
    """
    cases = {}
    key = ""

    for case in permanent_cases:
        cases[case] = 1.0
        key += f"{case.label}+"

    if leading_variable_case is not None:
        cases[leading_variable_case] = 1.0
        key += f"{leading_variable_case.label}+"

    for case in other_variable_cases:
        factor = PSI_FACTORS[VariableCategory(case.category)]["psi_0"]
        cases[case] = factor
        key += f"{factor}*{case.label}+"

    return cases, key[:-1]


def generate_sls_frequent_combination(
    permanent_cases: list[DesignLoadCase],
    leading_variable_case: DesignLoadCase | None,
    other_variable_cases: list[DesignLoadCase],
) -> tuple[dict[DesignLoadCase, float], str]:
    """
    Generate a frequent combination of load cases for serviceability limit state.

    Combination is generated according to equations 6.15 of EN 1990.
    The combination is normally used for reversible limit states.

    :param permanent_cases: A list of permanent load cases.
    :param leading_variable_case: The leading variable load case.
    :param other_variable_cases: A list of other variable load cases.
    """
    cases = {}
    key = ""

    for case in permanent_cases:
        cases[case] = 1.0
        key += f"{case.label}+"

    if leading_variable_case is not None:
        factor = PSI_FACTORS[VariableCategory(leading_variable_case.category)]["psi_1"]
        cases[leading_variable_case] = factor
        key += f"{factor}*{leading_variable_case.label}+"

    for case in other_variable_cases:
        factor = PSI_FACTORS[VariableCategory(case.category)]["psi_2"]
        cases[case] = factor
        key += f"{factor}*{case.label}+"

    return cases, key[:-1]


def generate_sls_quasipermanent_combination(
    permanent_cases: list[DesignLoadCase],
    leading_variable_case: DesignLoadCase | None,
    other_variable_cases: list[DesignLoadCase],
) -> tuple[dict[DesignLoadCase, float], str]:
    """
    Generate a quasipermanent combination of load cases for serviceability limit state.

    Combination is generated according to equations 6.16 of EN 1990.
    The combination is normally used for long-term effects and the appearance of the structure.

    :param permanent_cases: A list of permanent load cases.
    :param leading_variable_case: The leading variable load case.
    :param other_variable_cases: A list of other variable load cases.
    """
    cases = {}
    key = ""

    for case in permanent_cases:
        cases[case] = 1.0
        key += f"{case.label}+"

    if leading_variable_case is not None:
        factor = PSI_FACTORS[VariableCategory(leading_variable_case.category)]["psi_2"]
        cases[leading_variable_case] = factor
        key += f"{factor}*{leading_variable_case.label}+"

    for case in other_variable_cases:
        factor = PSI_FACTORS[VariableCategory(case.category)]["psi_2"]
        cases[case] = factor
        key += f"{factor}*{case.label}+"

    return cases, key[:-1]


def generate_uls_basic_combination(
    permanent_cases: list[DesignLoadCase],
    leading_variable_case: DesignLoadCase | None,
    other_variable_cases: list[DesignLoadCase],
) -> tuple[dict[DesignLoadCase, float], str]:
    """
    Generate a basic combination of load cases for ultimate limit state.

    Combination is generated according to equations 6.10 of EN 1990.

    :param permanent_cases: A list of permanent load cases.
    :param leading_variable_case: The leading variable load case.
    :param other_variable_cases: A list of other variable load cases.
    """
    cases = {}
    key = ""

    for case in permanent_cases:
        factor = GAMMA_VALUES["Set B"][case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        cases[case] = factor
        key += f"{factor}*{case.label}+"

    if leading_variable_case is not None:
        factor = GAMMA_VALUES["Set B"][leading_variable_case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        cases[leading_variable_case] = factor
        key += f"{factor}*{leading_variable_case.label}+"

    for case in other_variable_cases:
        gamma = GAMMA_VALUES["Set B"][case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        psi = PSI_FACTORS[VariableCategory(case.category)]["psi_0"]
        cases[case] = gamma * psi
        key += f"{gamma}*{psi}*{case.label}+"

    return cases, key[:-1]


def generate_uls_alternative_a_combination(
    permanent_cases: list[DesignLoadCase],
    leading_variable_case: DesignLoadCase | None,
    other_variable_cases: list[DesignLoadCase],
) -> tuple[dict[DesignLoadCase, float], str]:
    """
    Generate two alternative combinations of load cases for ultimate limit state.

    Combinations are generated according to equation 6.10a of EN 1990.

    :param permanent_cases: A list of permanent load cases.
    :param leading_variable_case: The leading variable load case.
    :param other_variable_cases: A list of other variable load cases.
    """
    cases = {}
    key = ""

    for case in permanent_cases:
        gamma = GAMMA_VALUES["Set B"][case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        cases[case] = gamma
        key += f"{gamma}*{case.label}+"

    if leading_variable_case is not None:
        gamma = GAMMA_VALUES["Set B"][leading_variable_case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        psi = PSI_FACTORS[VariableCategory(leading_variable_case.category)]["psi_0"]
        cases[leading_variable_case] = gamma * psi
        key += f"{gamma}*{psi}*{leading_variable_case.label}+"

    for case in other_variable_cases:
        gamma = GAMMA_VALUES["Set B"][case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        psi = PSI_FACTORS[VariableCategory(case.category)]["psi_0"]
        cases[case] = gamma * psi
        key += f"{gamma}*{psi}*{case.label}+"

    return cases, key[:-1]


def generate_uls_alternative_b_combination(
    permanent_cases: list[DesignLoadCase],
    leading_variable_case: DesignLoadCase | None,
    other_variable_cases: list[DesignLoadCase],
) -> tuple[dict[DesignLoadCase, float], str]:
    """
    Generate two alternative combinations of load cases for ultimate limit state.

    Combinations are generated according to equation 6.10b of EN 1990.

    :param permanent_cases: A list of permanent load cases.
    :param leading_variable_case: The leading variable load case.
    :param other_variable_cases: A list of other variable load cases.
    """
    cases = {}
    key = ""

    for case in permanent_cases:
        gamma = GAMMA_VALUES["Set B"][case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        cases[case] = gamma * XI
        key += f"{XI}*{gamma}*{case.label}+"

    if leading_variable_case is not None:
        gamma = GAMMA_VALUES["Set B"][leading_variable_case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        cases[leading_variable_case] = gamma
        key += f"{gamma}*{leading_variable_case.label}+"

    for case in other_variable_cases:
        gamma = GAMMA_VALUES["Set B"][case.load_type][
            LoadBehavior.UNFAVOURABLE
        ]  # TODO: Favourable?
        psi = PSI_FACTORS[VariableCategory(case.category)]["psi_0"]
        cases[case] = gamma * psi
        key += f"{gamma}*{psi}*{case.label}+"

    return cases, key[:-1]



from enum import IntEnum

from framesss.enums import CaseInsensitiveStrEnum


class WoodType(CaseInsensitiveStrEnum):
    """Enum for wood types."""

    SOLID_TIMBER = "solid timber"
    GLUED_LAMINATED_TIMBER = "glued laminated timber"
    LVL = "lvl"
    PLYWOOD = "plywood"
    OSB = "osb"
    PARTICLEBOARD = "particleboard"
    HARD_FIBREBOARD = "hard fibreboard"
    MEDIUM_FIBREBOARD = "medium fibreboard"
    MDF_FIBREBOARD = "mdf fibreboard"
    SOFT_FIBREBOARD = "soft fibreboard"


class JointType(CaseInsensitiveStrEnum):
    """Enum for joint types."""

    CONNECTION = "connection"
    PUNCHED_METAL_PLATE_FASTENERS = "punched metal plate fasteners"


class ServiceClass(IntEnum):
    """
    Enum for service classes.

    EN 1995-1-1, 2.3.1.3

    :cvar SC1: Service class 1 is characterised by a moisture content in the materials corresponding
               to a temperature of 20°C and the relative humidity of the surrounding air only
               exceeding 65 % for a few weeks per year. In service class 1 the average moisture content
               in most softwoods will not exceed 12 %.
    :cvar SC2: Service class 2 is characterised by a moisture content in the materials corresponding
               to a temperature of 20°C and the relative humidity of the surrounding air only
               exceeding 85 % for a few weeks per year. In service class 2 the average moisture content
               in most softwoods will not exceed 20 %.
    :cvar SC3: Service class 3 is characterised by climatic conditions leading to higher moisture
               contents than in service class 2.
    """

    SC1 = 1
    SC2 = 2
    SC3 = 3



from framesss.enums import CaseInsensitiveStrEnum


class Topography(CaseInsensitiveStrEnum):
    """Enumeration of topographies according to EN 1991-1-13."""

    WINDSWEPT = "windswept"
    NORMAL = "normal"
    SHELTERED = "sheltered"


class SnowZone(CaseInsensitiveStrEnum):
    """Enumeration of snow zones according to ČSN EN 1991-1-3."""

    I = "I"
    II = "II"
    III = "III"
    IV = "IV"
    V = "V"
    VI = "VI"
    VII = "VII"
    VIII = "VIII"
